##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" GenericSetup product interfaces
"""

from zope.interface import Attribute
from zope.interface import Interface
from zope.schema import Text
from zope.schema import TextLine


# Please note that these values may change. Always import
# the values from here instead of using the values directly.
BASE, EXTENSION = 1, 2
SKIPPED_FILES = ('CVS', '.svn', '_svn', '_darcs')
SKIPPED_SUFFIXES = ('~',)


class IPseudoInterface(Interface):

    """ API documentation;  not testable / enforceable.
    """


class ISetupEnviron(Interface):

    """Context for im- and export adapters.
    """

    def getLogger(name):
        """Get a logger with the specified name, creating it if necessary.
        """

    def shouldPurge():
        """When installing, should the existing setup be purged?
        """


class ISetupContext(ISetupEnviron):

    """ Context used for export / import plugins.
    """
    def getSite():
        """ Return the site object being configured / dumped.
        """

    def getSetupTool():
        """ Return the site object being configured / dumped.
        """

    def getEncoding():
        """ Get the encoding used for configuration data within the site.

        o Return None if the data should not be encoded.
        """

    def listNotes():
        """ Return notes recorded by this context.

        o Result a sequence of (component, message) tuples
        """

    def clearNotes():
        """ Clear all notes recorded by this context.
        """


class IImportContext(ISetupContext):

    def readDataFile(filename, subdir=None):
        """ Search the current configuration for the requested file.

        o 'filename' is the name (without path elements) of the file.

        o 'subdir' is an optional subdirectory;  if not supplied, search
          only the "root" directory.

        o Return the file contents as bytes, or None if the file cannot be
          found.
        """

    def getLastModified(path):
        """ Return the modification timestamp of the item at 'path'.

        o Result will be a DateTime instance.

        o Search profiles in the configuration in order.

        o If the context is filesystem based, return the 'stat' timestamp
          of the file / directory to which 'path' points.

        o If the context is ZODB-based, return the Zope modification time
          of the object to which 'path' points.

        o Return None if 'path' does not point to any object.
        """

    def isDirectory(path):
        """ Test whether path points to a directory / folder.

        o If the context is filesystem based, check that 'path' points to
          a subdirectory within the "root" directory.

        o If the context is ZODB-based, check that 'path' points to a
          "container" under the context's tool.

        o Return None if 'path' does not resolve;  otherwise, return a
          bool.
        """

    def listDirectory(path, skip=SKIPPED_FILES):
        """ List IDs of the contents of a  directory / folder.

        o Omit names in 'skip'.

        o If 'path' does not point to a directory / folder, return None.
        """


class IChunkableImportContext(IImportContext):

    def openDataFile(filename, subdir=None):
        """ Open a datafile for reading from the specified location.

        o 'filename' is the unqualified name of the file.

        o 'subdir', if passed, is a path to a subdirectory / folder in
          which to find the file;  if not passed, write the file to the
          "root" of the target.

        o Return a readable file-like object in binary mode that will return
          bytes when read from. The caller is responsible for calling 'close'
          on it.
        """


class IImportPlugin(IPseudoInterface):

    """ Signature for callables used to import portions of site configuration.
    """
    def __call__(context):
        """ Perform the setup step.

        o Return a message describing the work done.

        o 'context' must implement IImportContext.
        """


class IExportContext(ISetupContext):

    def writeDataFile(filename, text, content_type, subdir=None):
        """ Write data into the specified location.

        o 'filename' is the unqualified name of the file.

        o 'text' is the content of the file.

        o 'content_type' is the MIMEtype of the file.

        o 'subdir', if passed, is a path to a subdirectory / folder in
          which to write the file;  if not passed, write the file to the
          "root" of the target.
        """


class IChunkableExportContext(IExportContext):

    def openDataFile(filename, content_type, subdir=None):
        """ Open a datafile for writing into the specified location.

        o 'filename' is the unqualified name of the file.

        o 'content_type' is the MIMEtype of the file.

        o 'subdir', if passed, is a path to a subdirectory / folder in
          which to write the file;  if not passed, write the file to the
          "root" of the target.

        o Return a writeable file-like object in binary mode that expects
          bytes data. The caller is responsible for calling 'close' on it.
        """


class IExportPlugin(IPseudoInterface):

    """ Signature for callables used to export portions of site configuration.
    """
    def __call__(context):
        """ Write export data for the site wrapped by context.

        o Return a message describing the work done.

        o 'context' must implement IExportContext.  The plugin will use
          its 'writeDataFile' method for each file to be exported.
        """


class IStepRegistry(Interface):

    """ Base interface for step registries.
    """
    def listSteps():
        """ Return a sequence of IDs of registered steps.

        o Order is not significant.
        """

    def listStepMetadata():
        """ Return a sequence of mappings describing registered steps.

        o Mappings will be ordered alphabetically.
        """

    def getStepMetadata(key, default=None):
        """ Return a mapping of metadata for the step identified by 'key'.

        o Return 'default' if no such step is registered.

        o The 'handler' metadata is available via 'getStep'.
        """

    def generateXML():
        """ Return a round-trippable XML representation of the registry.

        o 'handler' values are serialized using their dotted names.
        """

    def parseXML(text):
        """ Parse 'text'.
        """


class IImportStepRegistry(IStepRegistry):

    """ API for import step registry.
    """
    def sortSteps():
        """ Return a sequence of registered step IDs

        o Sequence is sorted topologically by dependency, with the dependent
          steps *after* the steps they depend on.
        """

    def checkComplete():
        """ Return a sequence of (node, edge) tuples for unsatisifed deps.
        """

    def getStep(key, default=None):
        """ Return the IImportPlugin registered for 'key'.

        o Return 'default' if no such step is registered.
        """

    def registerStep(id, version=None, handler=None, dependencies=(),
                     title=None, description=None):
        """ Register a setup step.

        o 'id' is a unique name for this step,

        o 'version' is a string for comparing versions, it is preferred to
          be a yyyy/mm/dd-ii formatted string (date plus two-digit
          ordinal).  when comparing two version strings, the version with
          the lower sort order is considered the older version.

          - Newer versions of a step supplant older ones.

          - Attempting to register an older one after a newer one results
            in a KeyError.

          NOTE: The version argument is deprecated.

        o 'handler' should implement IImportPlugin.

        o 'dependencies' is a tuple of step ids which have to run before
          this step in order to be able to run at all. Registration of
          steps that have unmet dependencies are deferred until the
          dependencies have been registered.

        o 'title' is a one-line UI description for this step.
          If None, the first line of the documentation string of the handler
          is used, or the id if no docstring can be found.

        o 'description' is a one-line UI description for this step.
          If None, the remaining line of the documentation string of
          the handler is used, or default to ''.
        """


class IExportStepRegistry(IStepRegistry):

    """ API for export step registry.
    """
    def getStep(key, default=None):
        """ Return the IExportPlugin registered for 'key'.

        o Return 'default' if no such step is registered.
        """

    def registerStep(id, handler, title=None, description=None):
        """ Register an export step.

        o 'id' is the unique identifier for this step

        o 'handler' should implement IExportPlugin.

        o 'title' is a one-line UI description for this step.
          If None, the first line of the documentation string of the step
          is used, or the id if no docstring can be found.

        o 'description' is a one-line UI description for this step.
          If None, the remaining line of the documentation string of
          the step is used, or default to ''.
        """


class IToolsetRegistry(Interface):

    """ API for toolset registry.
    """
    def listForbiddenTools():
        """ Return a list of IDs of tools which must be removed, if present.
        """

    def addForbiddenTool(tool_id):
        """ Add 'tool_id' to the list of forbidden tools.

        o Raise KeyError if 'tool_id' is already in the list.

        o Raise ValueError if 'tool_id' is in the "required" list.
        """

    def listRequiredTools():
        """ Return a list of IDs of tools which must be present.
        """

    def getRequiredToolInfo(tool_id):
        """ Return a mapping describing a partiuclar required tool.

        o Keys include:

          'id' -- the ID of the tool

          'class' -- a dotted path to its class

        o Raise KeyError if 'tool_id' id not a known tool.
        """

    def listRequiredToolInfo():
        """ Return a list of IDs of tools which must be present.
        """

    def addRequiredTool(tool_id, dotted_name):
        """ Add a tool to our "required" list.

        o 'tool_id' is the tool's ID.

        o 'dotted_name' is a dotted (importable) name of the tool's class.

        o Raise KeyError if we have already registered a class for 'tool_id'.

        o Raise ValueError if 'tool_id' is in the "forbidden" list.
        """


class IProfileRegistry(Interface):

    """ API for profile registry.
    """
    def getProfileInfo(profile_id, for_=None):
        """ Return a mapping describing a registered filesystem profile.

        o Keys include:

          'id' -- the ID of the profile

          'title' -- its title

          'description' -- a textual description of the profile

          'path' -- a path to the profile on the filesystem.

          'product' -- the name of the product to which 'path' is
             relative (None for absolute paths).

          'type' -- either BASE or EXTENSION

        o 'for_', if passed, should be the interface specifying the "site
            type" for which the profile is relevant, e.g.
            Products.CMFCore.interfaces.ISiteRoot or
            Products.PluggableAuthService.interfaces.IPluggableAuthService.
            If 'None', list all profiles.
        """

    def listProfiles(for_=None):
        """ Return a list of IDs for registered profiles.

        o 'for_', if passed, should be the interface specifying the "site
            type" for which the profile is relevant, e.g.
            Products.CMFCore.interfaces.ISiteRoot or
            Products.PluggableAuthService.interfaces.IPluggableAuthService.
            If 'None', list all profiles.
        """

    def listProfileInfo(for_=None):
        """ Return a list of mappings describing registered profiles.

        o See 'getProfileInfo' for a description of the mappings' keys.

        o 'for_', if passed, should be the interface specifying the "site
            type" for which the profile is relevant, e.g.
            Products.CMFCore.interfaces.ISiteRoot or
            Products.PluggableAuthService.interfaces.IPluggableAuthService.
            If 'None', list all profiles.
        """

    def registerProfile(name, title, description, path, product=None,
                        profile_type=BASE, for_=None):
        """ Add a new profile to the registry.

        o If an existing profile is already registered for 'product:name',
          raise KeyError.

        o If 'product' is passed, then 'path' should be interpreted as
          relative to the corresponding product directory.

        o 'for_', if passed, should be the interface specifying the "site
          type" for which the profile is relevant, e.g.
          Products.CMFCore.interfaces.ISiteRoot or
          Products.PluggableAuthService.interfaces.IPluggableAuthService.
          If 'None', the profile might be used in any site.
        """


class ISetupTool(Interface):

    """ API for SetupTool.
    """

    def getEncoding():
        """ Get the encoding used for configuration data within the site.

        o Return None if the data should not be encoded.
        """

    def getBaselineContextID():
        """ Get the ID of the base profile for this configuration.
        """

    def setBaselineContext(context_id, encoding=None):
        """ Specify the base profile for this configuration.
        """

    def getExcludeGlobalSteps():
        """ Does this instance of the tool ignore globally-registered steps?
        """

    def setExcludeGlobalSteps(value):
        """ Specify whether to ignore globally-registered steps.

        'value' must be a boolean.
        """

    def applyContext(context, encoding=None):
        """ Update the tool from the supplied context, without modifying its
            "permanent" ID.
        """

    def getImportStepRegistry():
        """ Return the IImportStepRegistry for the tool.
        """

    def getExportStepRegistry():
        """ Return the IExportStepRegistry for the tool.
        """

    def getToolsetRegistry():
        """ Return the IToolsetRegistry for the tool.
        """

    def getProfileDependencyChain(profile_id):
        """Return a list of dependencies for a profile.

        The list is ordered by install order, with the requested profile as
        last item.
        """

    def runImportStepFromProfile(profile_id, step_id,
                                 run_dependencies=True, purge_old=None):
        """ Execute a given setup step from the given profile.

        o 'profile_id' must be a valid ID of a registered profile;
           otherwise, raise KeyError.

        o 'step_id' is the ID of the step to run.

        o If 'purge_old' is True, then run the step after purging any
          "old" setup first (this is the responsibility of the step,
          which must check the context we supply).

        o If 'run_dependencies' is True, then run any out-of-date
          dependency steps first.

        o Return a mapping, with keys:

          'steps' -- a sequence of IDs of the steps run.

          'messages' -- a dictionary holding messages returned from each
            step
        """

    def runAllImportStepsFromProfile(profile_id, purge_old=None,
                                     ignore_dependencies=False,
                                     blacklisted_steps=None):
        """ Run all setup steps for the given profile in dependency order.

        o 'profile_id' must be a valid ID of a registered profile;
           otherwise, raise KeyError.

        o If 'purge_old' is True, then run each step after purging any
          "old" setup first (this is the responsibility of the step,
          which must check the context we supply).

        o Unless 'ignore_dependencies' is true this will also import
          all profiles this profile depends on.

        o 'blacklisted_steps' can be a list of step-names that won't be
          executed. Use with special care and only for special cases.


        o Return a mapping, with keys:

          'steps' -- a sequence of IDs of the steps run.

          'messages' -- a dictionary holding messages returned from each
            step
        """

    def runExportStep(step_id):
        """ Generate a tarball containing artifacts from one export step.

        o 'step_id' identifies the export step.

        o Return a mapping, with keys:

          'steps' -- a sequence of IDs of the steps run.

          'messages' -- a dictionary holding messages returned from each
            step

          'tarball' -- the stringified tar-gz data.
        """

    def runAllExportSteps():
        """ Generate a tarball containing artifacts from all export steps.

        o Return a mapping, with keys:

          'steps' -- a sequence of IDs of the steps run.

          'messages' -- a dictionary holding messages returned from each
            step

          'tarball' -- the stringified tar-gz data.
        """

    def createSnapshot(snapshot_id):
        """ Create a snapshot folder using all steps.

        o 'snapshot_id' is the ID of the new folder.
        """

    def compareConfigurations(lhs_context, rhs_context,
                              missing_as_empty=False, ignore_whitespace=False):
        """ Compare two configurations.

        o 'lhs_context' and 'rhs_context' must implement IImportContext.

        o If 'missing_as_empty', then compare files not present as though
          they were zero-length;  otherwise, omit such files.

        o If 'ignore_whitespace', then suppress diffs due only to whitespace
          (c.f:  'diff -wbB')
        """

    def getProfileImportDate(profile_id):
        """ Return the last date an extension was imported.

        o The result will be a string, formated as IS0.
        """


class IWriteLogger(Interface):

    """Write methods used by the python logging Logger.
    """

    def debug(msg, *args, **kwargs):
        """Log 'msg % args' with severity 'DEBUG'.
        """

    def info(msg, *args, **kwargs):
        """Log 'msg % args' with severity 'INFO'.
        """

    def warning(msg, *args, **kwargs):
        """Log 'msg % args' with severity 'WARNING'.
        """

    def error(msg, *args, **kwargs):
        """Log 'msg % args' with severity 'ERROR'.
        """

    def exception(msg, *args):
        """Convenience method for logging an ERROR with exception information.
        """

    def critical(msg, *args, **kwargs):
        """Log 'msg % args' with severity 'CRITICAL'.
        """

    def log(level, msg, *args, **kwargs):
        """Log 'msg % args' with the integer severity 'level'.
        """


class INode(Interface):

    """Node im- and exporter.

    This is no generic DOM interface. It doesn't always provide a complete
    serialization of the object and its subobjects. If an adapter only
    provides 'INode' (and not 'IBody') it is likely to faithfully serialize
    the object. If, however, it also provides 'IBody', it is likely that its
    'node' serializes only an empty hull (without content or additional
    information such as properties) and expects that the missing parts are
    taken care of separately by the remaining 'IBody' infrastructure in a
    separate step. An essential example of this behavior is the default
    'INode' adapter for 'Folder' (and friends): its 'node' just specifies the
    folder's id and meta type, the folder's content is only serialized by
    'body'.
    """

    node = Text(description='Im- and export the object as a DOM node.')


class IBody(INode):

    """Body im- and exporter.
    """

    body = Text(description='Im- and export the object as a file body.')

    mime_type = TextLine(description='MIME type of the file body.')

    name = TextLine(description='Enforce this name for the file.')

    suffix = TextLine(description='Suffix for the file.')


class IFilesystemExporter(Interface):
    """ Plugin interface for site structure export.
    """
    def export(export_context, subdir, root=False):
        """ Export our 'context' using the API of 'export_context'.

        o 'export_context' must implement
          Products.GenericSupport.interfaces.IExportContext.

        o 'subdir', if passed, is the relative subdirectory containing our
          context within the site.

        o 'root', if true, indicates that the current context is the
          "root" of an import (this may be used to adjust paths when
          interacting with the context).
        """

    def listExportableItems():
        """ Return a sequence of the child items to be exported.

        o Each item in the returned sequence will be a tuple,
          (id, object, adapter) where adapter must implement
          IFilesystemExporter.
        """


class IFilesystemImporter(Interface):
    """ Plugin interface for site structure export.
    """
    def import_(import_context, subdir, root=False):
        """ Import our 'context' using the API of 'import_context'.

        o 'import_context' must implement
          Products.GenericSupport.interfaces.IImportContext.

        o 'subdir', if passed, is the relative subdirectory containing our
          context within the site.

        o 'root', if true, indicates that the current context is the
          "root" of an import (this may be used to adjust paths when
          interacting with the context).
        """


class IContentFactory(Interface):
    """ Adapter interface for factories specific to a container.
    """
    def __call__(id):
        """ Return a new instance, seated in the context under 'id'.
        """


class IContentFactoryName(Interface):
    """ Adapter interface for finding the name of the ICF for an object.
    """
    def __call__():
        """ Return a string, suitable for looking up an IContentFactory.

        o The string should allow finding a factory for our context's
          container which would create an "empty" instance of the same
          type as our context.
        """


class ICSVAware(Interface):
    """ Interface for objects which dump / load 'text/comma-separated-values'.
    """
    def getId():
        """ Return the Zope id of the object.
        """

    def as_csv():
        """ Return a string representing the object as CSV.
        """

    def put_csv(fd):
        """ Parse CSV and update the object.

        o 'fd' must be a file-like object whose 'read' method returns
          CSV text parseable by the 'csv.reader'.
        """


class IINIAware(Interface):
    """ Interface for objects which dump / load INI-format files..
    """
    def getId():
        """ Return the Zope id of the object.
        """

    def as_ini():
        """ Return a string representing the object as INI.
        """

    def put_ini(stream_or_text):
        """ Parse INI-formatted text and update the object.

        o 'stream_or_text' must be either a string, or else a stream
          directly parseable by ConfigParser.
        """


class IDAVAware(Interface):
    """ Interface for objects which handle their own FTP / DAV operations.
    """
    def getId():
        """ Return the Zope id of the object.
        """

    def manage_FTPget():
        """ Return a string representing the object as a file.
        """

    def PUT(REQUEST, RESPONSE):
        """ Parse file content and update the object.

        o 'REQUEST' will have a 'get' method, which will have the
          content object in its "BODY" key.  It will also have 'get_header'
          method, whose headers (e.g., "Content-Type") may affect the
          processing of the body.
        """


class IBeforeProfileImportEvent(Interface):
    """ An event which is fired before (part of) a profile is imported.
    """
    profile_id = Attribute("id of the profile to be imported "
                           "or None for non-profile imports.")

    steps = Attribute("list of steps that will be imported")

    full_import = Attribute("True if all steps will be imported")

    tool = Attribute("The tool which is performing the import")


class IProfileImportedEvent(Interface):
    """ An event which is fired when (part of) a profile is imported.
    """
    profile_id = Attribute("id of the imported profile")

    steps = Attribute("list of steps have been imported")

    full_import = Attribute("True if all steps are imported")

    tool = Attribute("The tool which is performing the import")


class IComponentsHandlerBlacklist(Interface):
    """ Interface for named utilities which can exclude specified interfaces
    from being handled by the components export and import handlers.
    """

    def getExcludedInterfaces():
        """ Return a sequence of interfaces.

        Objects providing any of the returned interfaces should be ignored by
        the export and import handlers.
        """


class IProfile(Interface):
    """ Named profile.
    """


class IImportStep(Interface):
    """ Named import step.
    """


class IExportStep(Interface):
    """ Named export step.
    """


class IUpgradeSteps(Interface):
    """ Named upgrade steps.
    """



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""OFSP export / import support.
"""

from OFS.interfaces import IFolder
from zope.component import adapts

from ..interfaces import ISetupEnviron
from ..utils import ObjectManagerHelpers
from ..utils import PropertyManagerHelpers
from ..utils import XMLAdapterBase


class FolderXMLAdapter(XMLAdapterBase, ObjectManagerHelpers,
                       PropertyManagerHelpers):

    """XML im- and exporter for Folder.
    """

    adapts(IFolder, ISetupEnviron)

    _LOGGER_ID = 'ofs'

    def _exportNode(self):
        """Export the object as a DOM node.
        """
        node = self._getObjectNode('object')
        node.appendChild(self._extractProperties())
        node.appendChild(self._extractObjects())

        self._logger.info('Folder exported.')
        return node

    def _importNode(self, node):
        """Import the object from the DOM node.
        """
        if self.environ.shouldPurge():
            self._purgeProperties()
            self._purgeObjects()

        self._initProperties(node)
        self._initObjects(node)

        self._logger.info('Folder imported.')

    def _exportBody(self):
        """Export the object as a file body.
        """
        if self.context.meta_type not in ('Folder', 'Folder (Ordered)'):
            return None

        return XMLAdapterBase._exportBody(self)

    body = property(_exportBody, XMLAdapterBase._importBody)



##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
""" Unit tests for GenericSetup tool.
"""

import os
import tempfile
import unittest
from io import BytesIO

import transaction
from AccessControl.Permissions import view
from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SecurityManagement import noSecurityManager
from AccessControl.users import UnrestrictedUser
from Acquisition import aq_base
from OFS.Folder import Folder
from zope.component import adapter
from zope.component import provideHandler
from zope.component.globalregistry import base as base_registry

from Products.GenericSetup import profile_registry

from ..context import TarballExportContext
from ..interfaces import IBeforeProfileImportEvent
from ..interfaces import IProfileImportedEvent
from ..testing import ExportImportZCMLLayer
from ..upgrade import UpgradeDepends
from ..upgrade import UpgradeStep
from ..upgrade import _registerNestedUpgradeStep
from ..upgrade import _registerUpgradeStep
from ..upgrade import listUpgradeSteps
from .common import BaseRegistryTests
from .common import DummyExportContext
from .common import DummyImportContext
from .common import FilesystemTestBase
from .common import TarballTester
from .common import _makeTestFile
from .conformance import ConformsToISetupTool
from .test_registry import _EMPTY_EXPORT_XML
from .test_registry import _EMPTY_IMPORT_XML
from .test_registry import _EMPTY_TOOLSET_XML
from .test_registry import _NORMAL_TOOLSET_XML
from .test_registry import _SINGLE_EXPORT_XML
from .test_registry import _SINGLE_IMPORT_XML
from .test_registry import ONE_FUNC
from .test_registry import IAnotherSite
from .test_registry import IDerivedSite
from .test_registry import ISite
from .test_zcml import dummy_upgrade


_before_import_events = []


@adapter(IBeforeProfileImportEvent)
def handleBeforeProfileImportEvent(event):
    _before_import_events.append(event)


_after_import_events = []


@adapter(IProfileImportedEvent)
def handleProfileImportedEvent(event):
    _after_import_events.append(event)


_METADATA_XML = """<?xml version="1.0"?>
<metadata>
  <version>1.0</version>
  <dependencies>
    <dependency>profile-other:bar</dependency>
  </dependencies>
</metadata>
"""
_DOUBLE_METADATA_XML = """<?xml version="1.0"?>
<metadata>
  <version>1.0</version>
  <dependencies>
    <dependency>profile-other:bar</dependency>
    <dependency>profile-other:ham</dependency>
  </dependencies>
</metadata>
"""
_PLAIN_METADATA_XML = """<?xml version="1.0"?>
<metadata>
  <version>1.0</version>
</metadata>
"""
_BROKEN_METADATA_XML = """<?xml version="1.0"?>
<metadata>
  <version>1.0</version>
  <dependencies>
    <dependency>profile-other:non-existing-profile</dependency>
  </dependencies>
</metadata>
"""


class SetupToolTests(FilesystemTestBase, TarballTester, ConformsToISetupTool):

    layer = ExportImportZCMLLayer

    _PROFILE_PATH = tempfile.mkdtemp(prefix='STT_test')
    _PROFILE_PATH2 = tempfile.mkdtemp(prefix='STT_test2')
    _PROFILE_PATH3 = tempfile.mkdtemp(prefix='STT_test3')

    def afterSetUp(self):
        from ..upgrade import _upgrade_registry
        _upgrade_registry.clear()
        profile_registry.clear()
        global _before_import_events
        global _after_import_events
        _before_import_events = []
        provideHandler(handleBeforeProfileImportEvent)
        _after_import_events = []
        provideHandler(handleProfileImportedEvent)

    def beforeTearDown(self):
        base_registry.unregisterHandler(handleBeforeProfileImportEvent)
        base_registry.unregisterHandler(handleProfileImportedEvent)
        FilesystemTestBase.beforeTearDown(self)
        from ..upgrade import _upgrade_registry
        profile_registry.clear()
        _upgrade_registry.clear()
        noSecurityManager()

    def _getTargetClass(self):
        from ..tool import SetupTool

        return SetupTool

    def _makeSite(self, title="Don't care"):

        site = Folder()
        site._setId('site')
        site.title = title

        self.app._setObject('site', site)
        self.app.acl_users.userFolderAddUser('admin', '', ['Manager'], [])
        newSecurityManager(None, self.app.acl_users.getUser('admin'))
        return self.app._getOb('site')

    def test_empty(self):

        tool = self._makeOne('setup_tool')

        self.assertEqual(tool.getBaselineContextID(), '')

        import_registry = tool.getImportStepRegistry()
        self.assertEqual(len(import_registry.listSteps()), 0)

        export_registry = tool.getExportStepRegistry()
        export_steps = export_registry.listSteps()
        self.assertEqual(len(export_steps), 0)

        toolset_registry = tool.getToolsetRegistry()
        self.assertEqual(len(toolset_registry.listForbiddenTools()), 0)
        self.assertEqual(len(toolset_registry.listRequiredTools()), 0)

    def test_getBaselineContextID(self):
        from ..tool import EXPORT_STEPS_XML
        from ..tool import IMPORT_STEPS_XML
        from ..tool import TOOLSET_XML

        tool = self._makeOne('setup_tool')

        self._makeFile(IMPORT_STEPS_XML, _EMPTY_IMPORT_XML)
        self._makeFile(EXPORT_STEPS_XML, _EMPTY_EXPORT_XML)
        self._makeFile(TOOLSET_XML, _EMPTY_TOOLSET_XML)

        profile_registry.registerProfile('foo', 'Foo', '', self._PROFILE_PATH)
        tool.setBaselineContext('profile-other:foo')

        self.assertEqual(tool.getBaselineContextID(), 'profile-other:foo')

    def test_setBaselineContext_invalid(self):

        tool = self._makeOne('setup_tool')

        self.assertRaises(KeyError, tool.setBaselineContext, 'profile-foo')

    def test_setBaselineContext_empty_string(self):

        tool = self._makeOne('setup_tool')

        self.assertRaises(KeyError, tool.setBaselineContext, '')

    def test_setBaselineContext(self):
        from ..tool import EXPORT_STEPS_XML
        from ..tool import IMPORT_STEPS_XML
        from ..tool import TOOLSET_XML

        tool = self._makeOne('setup_tool')
        tool.getExportStepRegistry().clear()

        self._makeFile(IMPORT_STEPS_XML, _SINGLE_IMPORT_XML)
        self._makeFile(EXPORT_STEPS_XML, _SINGLE_EXPORT_XML)
        self._makeFile(TOOLSET_XML, _NORMAL_TOOLSET_XML)

        profile_registry.registerProfile('foo', 'Foo', '', self._PROFILE_PATH)
        tool.setBaselineContext('profile-other:foo')

        self.assertEqual(tool.getBaselineContextID(), 'profile-other:foo')

        import_registry = tool.getImportStepRegistry()
        self.assertEqual(len(import_registry.listSteps()), 1)
        self.assertIn('one', import_registry.listSteps())
        info = import_registry.getStepMetadata('one')
        self.assertEqual(info['id'], 'one')
        self.assertEqual(info['title'], 'One Step')
        self.assertEqual(info['version'], '1')
        self.assertIn('One small step', info['description'])
        self.assertEqual(info['handler'],
                         'Products.GenericSetup.tests.test_registry.ONE_FUNC')

        self.assertEqual(import_registry.getStep('one'), ONE_FUNC)

        export_registry = tool.getExportStepRegistry()
        self.assertEqual(len(export_registry.listSteps()), 1)
        self.assertIn('one', import_registry.listSteps())
        info = export_registry.getStepMetadata('one')
        self.assertEqual(info['id'], 'one')
        self.assertEqual(info['title'], 'One Step')
        self.assertIn('One small step', info['description'])
        self.assertEqual(info['handler'],
                         'Products.GenericSetup.tests.test_registry.ONE_FUNC')

        self.assertEqual(export_registry.getStep('one'), ONE_FUNC)

    def test_runImportStepFromProfile_nonesuch(self):

        site = self._makeSite()

        tool = self._makeOne('setup_tool').__of__(site)

        self.assertRaises(KeyError, tool.runImportStepFromProfile, '',
                          'nonesuch')

    def test_runImportStepFromProfile_simple(self):

        TITLE = 'original title'
        site = self._makeSite(TITLE)

        tool = self._makeOne('setup_tool').__of__(site)

        registry = tool.getImportStepRegistry()
        registry.registerStep('simple', '1', _uppercaseSiteTitle)

        result = tool.runImportStepFromProfile('snapshot-dummy', 'simple')

        self.assertEqual(len(result['steps']), 1)

        self.assertEqual(result['steps'][0], 'simple')
        self.assertEqual(result['messages']['simple'], 'Uppercased title')

        self.assertEqual(site.title, TITLE.upper())

        self.assertEqual(len(_before_import_events), 1)
        self.assertEqual(_before_import_events[0].profile_id, 'snapshot-dummy')
        self.assertEqual(_before_import_events[0].steps, ['simple'])
        self.assertEqual(_before_import_events[0].full_import, False)

        self.assertEqual(len(_after_import_events), 1)
        self.assertEqual(_after_import_events[0].profile_id, 'snapshot-dummy')
        self.assertEqual(_after_import_events[0].steps, ['simple'])
        self.assertEqual(_after_import_events[0].full_import, False)

    def test_runImportStepFromProfile_dependencies(self):

        TITLE = 'original title'
        site = self._makeSite(TITLE)

        tool = self._makeOne('setup_tool').__of__(site)

        registry = tool.getImportStepRegistry()
        registry.registerStep('dependable', '1', _underscoreSiteTitle)
        registry.registerStep('dependent', '1', _uppercaseSiteTitle,
                              ('dependable', ))

        result = tool.runImportStepFromProfile('snapshot-dummy', 'dependent')

        self.assertEqual(len(result['steps']), 2)

        self.assertEqual(result['steps'][0], 'dependable')
        self.assertEqual(result['messages']['dependable'], 'Underscored title')

        self.assertEqual(result['steps'][1], 'dependent')
        self.assertEqual(result['messages']['dependent'], 'Uppercased title')
        self.assertEqual(site.title, TITLE.replace(' ', '_').upper())

        self.assertEqual(len(_before_import_events), 1)
        self.assertEqual(_before_import_events[0].profile_id, 'snapshot-dummy')
        self.assertEqual(_before_import_events[0].steps,
                         ['dependable', 'dependent'])
        self.assertEqual(_before_import_events[0].full_import, False)

        self.assertEqual(len(_after_import_events), 1)
        self.assertEqual(_after_import_events[0].profile_id, 'snapshot-dummy')
        self.assertEqual(_after_import_events[0].steps,
                         ['dependable', 'dependent'])
        self.assertEqual(_after_import_events[0].full_import, False)

    def test_runImportStepFromProfile_skip_dependencies(self):

        TITLE = 'original title'
        site = self._makeSite(TITLE)

        tool = self._makeOne('setup_tool').__of__(site)

        registry = tool.getImportStepRegistry()
        registry.registerStep('dependable', '1', _underscoreSiteTitle)
        registry.registerStep('dependent', '1', _uppercaseSiteTitle,
                              ('dependable', ))

        result = tool.runImportStepFromProfile('snapshot-dummy',
                                               'dependent',
                                               run_dependencies=False)

        self.assertEqual(len(result['steps']), 1)

        self.assertEqual(result['steps'][0], 'dependent')
        self.assertEqual(result['messages']['dependent'], 'Uppercased title')

        self.assertEqual(site.title, TITLE.upper())

        self.assertEqual(len(_before_import_events), 1)
        self.assertEqual(_before_import_events[0].profile_id, 'snapshot-dummy')
        self.assertEqual(_before_import_events[0].steps, ['dependent'])
        self.assertEqual(_before_import_events[0].full_import, False)

        self.assertEqual(len(_after_import_events), 1)
        self.assertEqual(_after_import_events[0].profile_id, 'snapshot-dummy')
        self.assertEqual(_after_import_events[0].steps, ['dependent'])
        self.assertEqual(_after_import_events[0].full_import, False)

    def test_runImportStepFromProfile_default_purge(self):

        site = self._makeSite()

        tool = self._makeOne('setup_tool').__of__(site)
        registry = tool.getImportStepRegistry()
        registry.registerStep('purging', '1', _purgeIfRequired)

        result = tool.runImportStepFromProfile('snapshot-dummy', 'purging')

        self.assertEqual(len(result['steps']), 1)
        self.assertEqual(result['steps'][0], 'purging')
        self.assertEqual(result['messages']['purging'], 'Purged')
        self.assertTrue(site.purged)

    def test_runImportStepFromProfile_explicit_purge(self):

        site = self._makeSite()

        tool = self._makeOne('setup_tool').__of__(site)
        registry = tool.getImportStepRegistry()
        registry.registerStep('purging', '1', _purgeIfRequired)

        result = tool.runImportStepFromProfile('snapshot-dummy',
                                               'purging',
                                               purge_old=True)

        self.assertEqual(len(result['steps']), 1)
        self.assertEqual(result['steps'][0], 'purging')
        self.assertEqual(result['messages']['purging'], 'Purged')
        self.assertTrue(site.purged)

    def test_runImportStepFromProfile_skip_purge(self):

        site = self._makeSite()

        tool = self._makeOne('setup_tool').__of__(site)
        registry = tool.getImportStepRegistry()
        registry.registerStep('purging', '1', _purgeIfRequired)

        result = tool.runImportStepFromProfile('snapshot-dummy',
                                               'purging',
                                               purge_old=False)

        self.assertEqual(len(result['steps']), 1)
        self.assertEqual(result['steps'][0], 'purging')
        self.assertEqual(result['messages']['purging'], 'Unpurged')
        self.assertFalse(site.purged)

    def test_runImportStepFromProfile_consistent_context(self):

        site = self._makeSite()

        tool = self._makeOne('setup_tool').__of__(site)

        registry = tool.getImportStepRegistry()
        registry.registerStep('purging', '1', _purgeIfRequired)
        registry.registerStep('dependent', '1', _uppercaseSiteTitle,
                              ('purging', ))

        tool.runImportStepFromProfile('snapshot-dummy',
                                      'dependent',
                                      purge_old=False)
        self.assertFalse(site.purged)

    def test_runAllImportStepsFromProfile_empty(self):

        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)

        result = tool.runAllImportStepsFromProfile('snapshot-dummy')

        self.assertEqual(len(result['steps']), 3)

    def test_runAllImportStepsFromProfile_inquicksuccession(self):
        """
        This test provokes an issue that only appears in testing.
        There it can happen that profiles get run multiple times within
        a second. As of 1.6.3, genericsetup does not handle this.
        """

        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)

        tool.runAllImportStepsFromProfile('snapshot-dummy')
        tool.runAllImportStepsFromProfile('snapshot-dummy')
        # For good measurement
        tool.runAllImportStepsFromProfile('snapshot-dummy')

        self.assertTrue("No exception thrown")

    def test_runAllImportStepsFromProfile_sorted_default_purge(self):

        TITLE = 'original title'
        PROFILE_ID = 'snapshot-testing'
        site = self._makeSite(TITLE)
        tool = self._makeOne('setup_tool').__of__(site)
        tool._exclude_global_steps = True

        registry = tool.getImportStepRegistry()
        registry.registerStep('dependable', '1', _underscoreSiteTitle,
                              ('purging', ))
        registry.registerStep('dependent', '1', _uppercaseSiteTitle,
                              ('dependable', ))
        registry.registerStep('purging', '1', _purgeIfRequired)

        result = tool.runAllImportStepsFromProfile(PROFILE_ID)

        self.assertEqual(len(result['steps']), 3)

        self.assertEqual(result['steps'][0], 'purging')
        self.assertEqual(result['messages']['purging'], 'Purged')

        self.assertEqual(result['steps'][1], 'dependable')
        self.assertEqual(result['messages']['dependable'], 'Underscored title')

        self.assertEqual(result['steps'][2], 'dependent')
        self.assertEqual(result['messages']['dependent'], 'Uppercased title')

        self.assertEqual(site.title, TITLE.replace(' ', '_').upper())
        self.assertTrue(site.purged)

        prefix = 'import-all-%s' % PROFILE_ID
        logged = [x for x in tool.objectIds('File') if x.startswith(prefix)]
        self.assertEqual(len(logged), 1)

    def check_restricted_access(self, obj):
        # For most objects that we create, we do not want ordinary users to
        # see it, also not when they have View permission on a higher level.
        rop_info = obj.rolesOfPermission(view)
        allowed_roles = sorted([x['name'] for x in rop_info if x['selected']])
        self.assertEqual(allowed_roles, ['Manager', 'Owner'])
        self.assertFalse(obj.acquiredRolesAreUsedBy(view))

    def test_runAllImportStepsFromProfile_unicode_id_creates_reports(self):

        TITLE = 'original title'
        PROFILE_ID = 'snapshot-testing'
        site = self._makeSite(TITLE)
        tool = self._makeOne('setup_tool').__of__(site)

        registry = tool.getImportStepRegistry()
        registry.registerStep('dependable', '1', _underscoreSiteTitle,
                              ('purging', ))
        registry.registerStep('dependent', '1', _uppercaseSiteTitle,
                              ('dependable', ))
        registry.registerStep('purging', '1', _purgeIfRequired)

        tool.runAllImportStepsFromProfile(PROFILE_ID)

        prefix = str('import-all-%s' % PROFILE_ID)
        logged = [x for x in tool.objectIds('File') if x.startswith(prefix)]
        self.assertEqual(len(logged), 1)

        # Check acess restriction on log files
        logged = [x for x in tool.objectIds('File')]
        for file_id in logged:
            file_ob = tool._getOb(file_id)
            self.check_restricted_access(file_ob)

    def test_runAllImportStepsFromProfile_sorted_explicit_purge(self):

        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)
        tool._exclude_global_steps = True

        registry = tool.getImportStepRegistry()
        registry.registerStep('dependable', '1', _underscoreSiteTitle,
                              ('purging', ))
        registry.registerStep('dependent', '1', _uppercaseSiteTitle,
                              ('dependable', ))
        registry.registerStep('purging', '1', _purgeIfRequired)

        result = tool.runAllImportStepsFromProfile('snapshot-dummy',
                                                   purge_old=True)

        self.assertEqual(len(result['steps']), 3)

        self.assertEqual(result['steps'][0], 'purging')
        self.assertEqual(result['messages']['purging'], 'Purged')

        self.assertEqual(result['steps'][1], 'dependable')
        self.assertEqual(result['steps'][2], 'dependent')
        self.assertTrue(site.purged)

    def test_runAllImportStepsFromProfile_sorted_skip_purge(self):

        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)
        tool._exclude_global_steps = True

        registry = tool.getImportStepRegistry()
        registry.registerStep('dependable', '1', _underscoreSiteTitle,
                              ('purging', ))
        registry.registerStep('dependent', '1', _uppercaseSiteTitle,
                              ('dependable', ))
        registry.registerStep('purging', '1', _purgeIfRequired)

        result = tool.runAllImportStepsFromProfile('snapshot-dummy',
                                                   purge_old=False)

        self.assertEqual(len(result['steps']), 3)

        self.assertEqual(result['steps'][0], 'purging')
        self.assertEqual(result['messages']['purging'], 'Unpurged')

        self.assertEqual(result['steps'][1], 'dependable')
        self.assertEqual(result['steps'][2], 'dependent')
        self.assertFalse(site.purged)

    def test_runAllImportStepsFromProfile_without_depends(self):
        from ..metadata import METADATA_XML

        self._makeFile(METADATA_XML, _METADATA_XML)

        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)

        profile_registry.registerProfile('foo', 'Foo', '', self._PROFILE_PATH)

        _imported = []

        def applyContext(context):
            _imported.append(context._profile_path)

        tool.applyContext = applyContext
        tool.runAllImportStepsFromProfile('profile-other:foo',
                                          ignore_dependencies=True)
        self.assertEqual(_imported, [self._PROFILE_PATH])

    def test_runAllImportStepsFromProfile_with_depends(self):
        from ..metadata import METADATA_XML

        self._makeFile(METADATA_XML, _METADATA_XML)

        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)

        profile_registry.registerProfile('foo', 'Foo', '', self._PROFILE_PATH)
        profile_registry.registerProfile('bar', 'Bar', '', self._PROFILE_PATH2)

        _imported = []

        def applyContext(context):
            _imported.append(context._profile_path)

        tool.applyContext = applyContext
        tool.runAllImportStepsFromProfile('profile-other:foo',
                                          ignore_dependencies=False)
        self.assertEqual(_imported, [self._PROFILE_PATH2, self._PROFILE_PATH])

    def _setup_dependency_strategy_test_tool(self):
        # If we add a dependency profile in our metadata.xml, and this
        # dependency was already applied, then we do not need to apply
        # it yet again.  Once is quite enough, thank you.  Running any
        # upgrade steps would be nice though.  There are options.
        # Setup a tool and profiles for testing dependency strategies.
        from ..interfaces import EXTENSION
        from ..metadata import METADATA_XML
        self._makeFile(METADATA_XML, _DOUBLE_METADATA_XML)
        _makeTestFile(METADATA_XML, self._PROFILE_PATH2, _PLAIN_METADATA_XML)
        _makeTestFile(METADATA_XML, self._PROFILE_PATH3, _PLAIN_METADATA_XML)
        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)

        # Register main profile and two dependency profiles.
        profile_registry.registerProfile('foo',
                                         'Foo',
                                         '',
                                         self._PROFILE_PATH,
                                         profile_type=EXTENSION)
        profile_registry.registerProfile('bar',
                                         'Bar',
                                         '',
                                         self._PROFILE_PATH2,
                                         profile_type=EXTENSION)
        profile_registry.registerProfile('ham',
                                         'Ham',
                                         '',
                                         self._PROFILE_PATH3,
                                         profile_type=EXTENSION)

        # Apply the second profile.
        tool.runAllImportStepsFromProfile('profile-other:bar')

        # Register an upgrade step.  Note that applying this step will
        # set the profile version to 1.1, even though the metadata of
        # the profile really says 1.0.  We will use this to check
        # whether the upgrade step has been applied (version is 1.1)
        # or the full profile has been applied (version is 1.0).
        step_bar = UpgradeStep("Upgrade", "other:bar", '1.0', '1.1', '',
                               dummy_upgrade, None, "1")
        _registerUpgradeStep(step_bar)
        # And another one.
        step_ham = UpgradeStep("Upgrade", "other:ham", '1.0', '1.1', '',
                               dummy_upgrade, None, "1")
        _registerUpgradeStep(step_ham)

        # Gather list of imported profiles.
        tool._imported = []

        def applyContext(context):
            tool._imported.append(context._profile_path)

        tool.applyContext = applyContext

        return tool

    def test_runAllImportStepsFromProfile_with_default_strategy(self):
        # Default strategy: apply new profiles, upgrade old profiles.
        tool = self._setup_dependency_strategy_test_tool()

        # Run the main profile.
        tool.runAllImportStepsFromProfile('profile-other:foo')
        # The main and third profile have been applied.
        self.assertEqual(tool._imported,
                         [self._PROFILE_PATH3, self._PROFILE_PATH])
        # The upgrade step of the second profile has been applied,
        # pushing it to version 1.1.
        self.assertEqual(tool.getLastVersionForProfile('other:bar'),
                         ('1', '1'))
        # Third profile is at 1.0.
        self.assertEqual(tool.getLastVersionForProfile('other:ham'),
                         ('1', '0'))

    def test_runAllImportStepsFromProfile_with_reapply_strategy(self):
        # You can choose the old behavior of always applying the
        # dependencies.  This ignores any upgrade steps.
        tool = self._setup_dependency_strategy_test_tool()

        # Run the main profile.
        from ..tool import DEPENDENCY_STRATEGY_REAPPLY
        tool.runAllImportStepsFromProfile(
            'profile-other:foo',
            dependency_strategy=DEPENDENCY_STRATEGY_REAPPLY)
        # All three profiles have been applied.
        self.assertEqual(
            tool._imported,
            [self._PROFILE_PATH2, self._PROFILE_PATH3, self._PROFILE_PATH])
        self.assertEqual(tool.getLastVersionForProfile('other:bar'),
                         ('1', '0'))
        self.assertEqual(tool.getLastVersionForProfile('other:ham'),
                         ('1', '0'))

    def test_runAllImportStepsFromProfile_with_new_strategy(self):
        # You can choose to be happy with any applied version and
        # ignore any upgrade steps.
        tool = self._setup_dependency_strategy_test_tool()

        # Run the main profile.
        from ..tool import DEPENDENCY_STRATEGY_NEW
        tool.runAllImportStepsFromProfile(
            'profile-other:foo', dependency_strategy=DEPENDENCY_STRATEGY_NEW)
        # The main and third profile have been applied.
        self.assertEqual(tool._imported,
                         [self._PROFILE_PATH3, self._PROFILE_PATH])
        # Second profile stays at 1.0.
        self.assertEqual(tool.getLastVersionForProfile('other:bar'),
                         ('1', '0'))
        self.assertEqual(tool.getLastVersionForProfile('other:ham'),
                         ('1', '0'))

    def test_runAllImportStepsFromProfile_with_ignore_strategy(self):
        # You can choose to be ignore all dependency profiles.
        tool = self._setup_dependency_strategy_test_tool()

        # Run the main profile.
        from ..tool import DEPENDENCY_STRATEGY_IGNORE
        tool.runAllImportStepsFromProfile(
            'profile-other:foo',
            dependency_strategy=DEPENDENCY_STRATEGY_IGNORE)
        # Only the main profile has been applied.
        self.assertEqual(tool._imported, [self._PROFILE_PATH])
        # Second profile stays at 1.0.
        self.assertEqual(tool.getLastVersionForProfile('other:bar'),
                         ('1', '0'))
        # Third profile is not applied.
        self.assertEqual(tool.getLastVersionForProfile('other:ham'),
                         ('unknown'))

    def test_runAllImportStepsFromProfile_unknown_strategy(self):
        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)
        profile_registry.registerProfile('foo', 'Foo', '', self._PROFILE_PATH)
        self.assertRaises(ValueError,
                          tool.runAllImportStepsFromProfile,
                          'profile-other:foo',
                          dependency_strategy='random')

    def test_runAllImportStepsFromProfile_set_last_profile_version(self):
        from ..metadata import METADATA_XML

        self._makeFile(METADATA_XML, _METADATA_XML)

        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)

        profile_registry.registerProfile('foo', 'Foo', '', self._PROFILE_PATH)

        # test initial states
        profile_id = "other:foo"
        self.assertEqual(tool.getVersionForProfile(profile_id), '1.0')
        self.assertEqual(tool.getLastVersionForProfile(profile_id), 'unknown')

        # run all imports steps
        tool.runAllImportStepsFromProfile('profile-other:foo',
                                          ignore_dependencies=True)

        # events.handleProfileImportedEvent should set last profile version
        self.assertEqual(tool.getLastVersionForProfile(profile_id), ('1', '0'))

    def test_runAllImportStepsFromProfile_step_registration_with_depends(self):
        from ..metadata import METADATA_XML

        self._makeFile(METADATA_XML, _METADATA_XML)

        _IMPORT_STEPS_XML = """<?xml version="1.0"?>
<import-steps>
 <import-step id="one"
             version="1"
             handler="Products.GenericSetup.tests.common.dummy_handler"
             title="One Step">
  One small step
 </import-step>
</import-steps>
"""
        self._makeFile('import_steps.xml', _IMPORT_STEPS_XML)

        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)

        profile_registry.registerProfile('foo', 'Foo', '', self._PROFILE_PATH)
        profile_registry.registerProfile('bar', 'Bar', '', self._PROFILE_PATH2)

        result = tool.runAllImportStepsFromProfile('profile-other:foo',
                                                   ignore_dependencies=False)

        # ensure the additional step on foo was imported
        self.assertIn('one', result['steps'])

    def test_runAllImportStepsFromProfile_skipStep(self):

        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)
        result = tool.runAllImportStepsFromProfile(
            'snapshot-dummy',
            blacklisted_steps=['toolset'],
        )

        self.assertEqual((result['messages']['toolset']), 'step skipped')

    def test_runAllImportStepsFromProfile_with_base_profile(self):
        # Applying a base profile should clear the profile upgrade
        # versions.
        from ..interfaces import BASE
        from ..interfaces import EXTENSION
        from ..metadata import METADATA_XML
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        self._makeFile(METADATA_XML, _METADATA_XML)
        _makeTestFile(METADATA_XML, self._PROFILE_PATH2, _PLAIN_METADATA_XML)
        _makeTestFile(METADATA_XML, self._PROFILE_PATH3, _PLAIN_METADATA_XML)

        # Register a base and two extension profile.  The base profile
        # 'foo' has a dependency 'bar'.  This might not make sense,
        # but it will serve to check that we clear the profile
        # versions right before we apply the base profile, which means
        # right after any dependency profiles.
        profile_registry.registerProfile('foo',
                                         'Foo',
                                         '',
                                         self._PROFILE_PATH,
                                         profile_type=BASE)
        profile_registry.registerProfile('bar',
                                         'Bar',
                                         '',
                                         self._PROFILE_PATH2,
                                         profile_type=EXTENSION)
        profile_registry.registerProfile('ham',
                                         'Ham',
                                         '',
                                         self._PROFILE_PATH3,
                                         profile_type=EXTENSION)
        # Apply the extension profile.
        tool.runAllImportStepsFromProfile('profile-other:ham')
        self.assertEqual(tool._profile_upgrade_versions,
                         {'other:ham': ('1', '0')})
        # Apply the base profile.
        tool.runAllImportStepsFromProfile('profile-other:foo')
        self.assertEqual(tool._profile_upgrade_versions,
                         {'other:foo': ('1', '0')})

    def test_runAllImportStepsFromProfile_with_unknown_pre_handler(self):
        # Registering already fails.
        self.assertRaises(
            ValueError,
            profile_registry.registerProfile,
            'foo',
            'Foo',
            '',
            self._PROFILE_PATH,
            pre_handler='Products.GenericSetup.tests.test_tool.foo_handler')

    def test_runAllImportStepsFromProfile_with_unknown_post_handler(self):
        # Registering already fails.
        self.assertRaises(
            ValueError,
            profile_registry.registerProfile,
            'foo',
            'Foo',
            '',
            self._PROFILE_PATH,
            post_handler='Products.GenericSetup.tests.test_tool.foo_handler')

    def test_runAllImportStepsFromProfile_pre_post_handlers_dotted_names(self):
        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)
        profile_registry.registerProfile(
            'foo',
            'Foo',
            '',
            self._PROFILE_PATH,
            pre_handler='Products.GenericSetup.tests.test_tool.pre_handler',
            post_handler='Products.GenericSetup.tests.test_tool.post_handler')
        tool.runAllImportStepsFromProfile('profile-other:foo')
        self.assertEqual(tool.pre_handler_called, 1)
        self.assertEqual(tool.post_handler_called, 1)
        tool.runAllImportStepsFromProfile('profile-other:foo')
        self.assertEqual(tool.pre_handler_called, 2)
        self.assertEqual(tool.post_handler_called, 2)

    def test_runAllImportStepsFromProfile_pre_post_handlers_functions(self):
        # When you register a profile with pre/post handlers in zcml, you do
        # not get dotted names (strings) but an actual function.
        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)
        profile_registry.registerProfile('foo',
                                         'Foo',
                                         '',
                                         self._PROFILE_PATH,
                                         pre_handler=pre_handler,
                                         post_handler=post_handler)
        tool.runAllImportStepsFromProfile('profile-other:foo')
        self.assertEqual(tool.pre_handler_called, 1)
        self.assertEqual(tool.post_handler_called, 1)
        tool.runAllImportStepsFromProfile('profile-other:foo')
        self.assertEqual(tool.pre_handler_called, 2)
        self.assertEqual(tool.post_handler_called, 2)

    def test_runExportStep_nonesuch(self):
        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)

        self.assertRaises(ValueError, tool.runExportStep, 'nonesuch')

    def test_runExportStep_step_registry_empty(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool

        result = tool.runExportStep('step_registries')

        self.assertEqual(len(result['steps']), 1)
        self.assertEqual(result['steps'][0], 'step_registries')
        self.assertEqual(result['messages']['step_registries'], None)

    def test_runExportStep_step_registry_default(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        tool._import_registry.registerStep('foo', handler='foo.bar')
        steps = 'Products.GenericSetup.tool.exportStepRegistries'
        tool._export_registry.registerStep('step_registries', steps,
                                           'Export import / export steps.')

        result = tool.runExportStep('step_registries')

        self.assertEqual(len(result['steps']), 1)
        self.assertEqual(result['steps'][0], 'step_registries')
        self.assertEqual(result['messages']['step_registries'], None)
        fileish = BytesIO(result['tarball'])

        self._verifyTarballContents(fileish,
                                    ['import_steps.xml', 'export_steps.xml'])
        self._verifyTarballEntryXML(fileish, 'import_steps.xml',
                                    _DEFAULT_STEP_REGISTRIES_IMPORT_XML)
        self._verifyTarballEntryXML(fileish, 'export_steps.xml',
                                    _DEFAULT_STEP_REGISTRIES_EXPORT_XML)

    def test_runAllExportSteps_empty(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool

        result = tool.runAllExportSteps()

        self.assertEqual(
            sorted(result['steps']),
            ['componentregistry', 'rolemap', 'step_registries', 'toolset'])
        self.assertEqual(result['messages']['step_registries'], None)

    def test_runAllExportSteps_default(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        tool._import_registry.registerStep('foo', handler='foo.bar')
        steps = 'Products.GenericSetup.tool.exportStepRegistries'
        tool._export_registry.registerStep('step_registries', steps,
                                           'Export import / export steps.')

        result = tool.runAllExportSteps()

        self.assertEqual(
            sorted(result['steps']),
            ['componentregistry', 'rolemap', 'step_registries', 'toolset'])
        self.assertEqual(result['messages']['step_registries'], None)
        fileish = BytesIO(result['tarball'])

        self._verifyTarballContents(fileish, [
            'import_steps.xml', 'export_steps.xml', 'rolemap.xml',
            'toolset.xml'
        ])
        self._verifyTarballEntryXML(fileish, 'import_steps.xml',
                                    _DEFAULT_STEP_REGISTRIES_IMPORT_XML)
        self._verifyTarballEntryXML(fileish, 'export_steps.xml',
                                    _DEFAULT_STEP_REGISTRIES_EXPORT_XML)

    def test_runAllExportSteps_extras(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        steps = 'Products.GenericSetup.tool.exportStepRegistries'
        tool._export_registry.registerStep('step_registries', steps,
                                           'Export import / export steps.')

        import_reg = tool.getImportStepRegistry()
        import_reg.registerStep('dependable', '1', _underscoreSiteTitle,
                                ('purging', ))
        import_reg.registerStep('dependent', '1', _uppercaseSiteTitle,
                                ('dependable', ))
        import_reg.registerStep('purging', '1', _purgeIfRequired)

        export_reg = tool.getExportStepRegistry()
        export_reg.registerStep('properties', _exportPropertiesINI)

        result = tool.runAllExportSteps()

        self.assertEqual(len(result['steps']), 5)
        self.assertEqual(sorted(result['steps']), [
            'componentregistry', 'properties', 'rolemap', 'step_registries',
            'toolset'
        ])

        self.assertEqual(result['messages']['properties'],
                         'Exported properties')
        self.assertEqual(result['messages']['step_registries'], None)

        fileish = BytesIO(result['tarball'])

        self._verifyTarballContents(fileish, [
            'import_steps.xml', 'export_steps.xml', 'properties.ini',
            'rolemap.xml', 'toolset.xml'
        ])
        self._verifyTarballEntryXML(fileish, 'import_steps.xml',
                                    _EXTRAS_STEP_REGISTRIES_IMPORT_XML)
        self._verifyTarballEntryXML(fileish, 'export_steps.xml',
                                    _EXTRAS_STEP_REGISTRIES_EXPORT_XML)
        ini_string = _PROPERTIES_INI % site.title
        self._verifyTarballEntry(fileish, 'properties.ini',
                                 ini_string.encode('utf-8'))

    def test_manage_importTarball(self):
        # Tests for importing a tarball with GenericSetup files.
        # We are especially interested to see if old settings get purged.
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        # We need to be Manager to see the result of calling
        # manage_importTarball.
        newSecurityManager(None, UnrestrictedUser('root', '', ['Manager'], ''))

        ROLEMAP_XML = """<?xml version="1.0"?>
<rolemap>
  <roles>
    <role name="%s" />
  </roles>
  <permissions />
</rolemap>
"""

        def rolemap_tarball(name):
            # Create a tarball archive with rolemap.xml containing 'name' as
            # role.
            context = TarballExportContext(tool)
            contents = ROLEMAP_XML % name
            if isinstance(contents, str):
                contents = contents.encode('utf-8')
            context.writeDataFile('rolemap.xml', contents, 'text/xml')
            return context.getArchive()

        # Import first role.
        tool.manage_importTarball(rolemap_tarball('First'))
        self.assertIn('First', site.valid_roles())

        # Import second role.
        tool.manage_importTarball(rolemap_tarball('Second'))
        self.assertIn('Second', site.valid_roles())
        # The first role has been purged, because that is the default.
        self.assertNotIn('First', site.valid_roles())
        # A few standard roles are never removed, probably because they are
        # defined one level higher.
        self.assertIn('Anonymous', site.valid_roles())
        self.assertIn('Authenticated', site.valid_roles())
        self.assertIn('Manager', site.valid_roles())
        self.assertIn('Owner', site.valid_roles())

        # Import third role in non-purge mode.
        tool.manage_importTarball(rolemap_tarball('Third'), purge_old=False)
        self.assertIn('Third', site.valid_roles())
        # The second role is still there.
        self.assertIn('Second', site.valid_roles())

        # When you use the form, and uncheck the purge_old checkbox, then the
        # browser does not send the purge_old parameter in the request.  To
        # work around this, the form always passes a hidden 'submitted'
        # parameter.
        # Import fourth role in non-purge mode with a form submit.
        tool.manage_importTarball(rolemap_tarball('Fourth'), submitted='yes')
        self.assertIn('Fourth', site.valid_roles())
        # The other roles are still there.
        self.assertIn('Second', site.valid_roles())
        self.assertIn('Third', site.valid_roles())
        self.assertIn('Manager', site.valid_roles())

    def test_createSnapshot_default(self):
        _EXPECTED = [
            ('import_steps.xml', _DEFAULT_STEP_REGISTRIES_IMPORT_XML),
            ('export_steps.xml', _DEFAULT_STEP_REGISTRIES_EXPORT_XML),
            ('rolemap.xml', 'dummy'),
            ('toolset.xml', 'dummy'),
        ]

        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        tool._import_registry.registerStep('foo', handler='foo.bar')
        tool._export_registry.registerStep(
            'step_registries',
            'Products.GenericSetup.tool.exportStepRegistries',
            'Export import / export steps.')

        self.assertEqual(len(tool.listSnapshotInfo()), 0)

        result = tool.createSnapshot('default')

        self.assertEqual(
            sorted(result['steps']),
            ['componentregistry', 'rolemap', 'step_registries', 'toolset'])
        self.assertEqual(result['messages']['step_registries'], None)

        snapshot = result['snapshot']

        self.assertEqual(len(snapshot.objectIds()), len(_EXPECTED))

        for id in [x[0] for x in _EXPECTED]:
            self.assertIn(id, snapshot.objectIds())

        def normalize_xml(xml):
            # using this might mask a real problem on windows, but so far the
            # different newlines just caused problems in this test
            lines = [line.strip() for line in xml.splitlines() if line.strip()]
            return ' '.join(lines)

        fileobj = snapshot._getOb('import_steps.xml')
        self.assertEqual(normalize_xml(fileobj.read()),
                         normalize_xml(_DEFAULT_STEP_REGISTRIES_IMPORT_XML))

        fileobj = snapshot._getOb('export_steps.xml')
        self.assertEqual(normalize_xml(fileobj.read()),
                         normalize_xml(_DEFAULT_STEP_REGISTRIES_EXPORT_XML))

        self.assertEqual(len(tool.listSnapshotInfo()), 1)

        info = tool.listSnapshotInfo()[0]

        self.assertEqual(info['id'], 'default')
        self.assertEqual(info['title'], 'default')

        # Check access restriction on snapshot files and folders
        self.check_restricted_access(tool.snapshots)
        self.check_restricted_access(snapshot)
        for obj in snapshot.objectValues():
            self.check_restricted_access(obj)
            if hasattr(aq_base(obj), 'objectValues'):
                for child in obj.objectValues():
                    self.check_restricted_access(child)

    def test_applyContext(self):
        from ..tool import EXPORT_STEPS_XML
        from ..tool import IMPORT_STEPS_XML
        from ..tool import TOOLSET_XML

        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)
        tool.getImportStepRegistry().clear()
        tool.getExportStepRegistry().clear()
        tool.getToolsetRegistry().clear()

        context = DummyImportContext(site, tool=tool)
        context._files[IMPORT_STEPS_XML] = _SINGLE_IMPORT_XML
        context._files[EXPORT_STEPS_XML] = _SINGLE_EXPORT_XML
        context._files[TOOLSET_XML] = _NORMAL_TOOLSET_XML

        tool.applyContext(context)

        import_registry = tool.getImportStepRegistry()
        self.assertEqual(len(import_registry.listSteps()), 1)
        self.assertIn('one', import_registry.listSteps())
        info = import_registry.getStepMetadata('one')

        self.assertEqual(info['id'], 'one')
        self.assertEqual(info['title'], 'One Step')
        self.assertEqual(info['version'], '1')
        self.assertIn('One small step', info['description'])
        self.assertEqual(info['handler'],
                         'Products.GenericSetup.tests.test_registry.ONE_FUNC')

        self.assertEqual(import_registry.getStep('one'), ONE_FUNC)

        export_registry = tool.getExportStepRegistry()
        self.assertEqual(len(export_registry.listSteps()), 1)
        self.assertIn('one', import_registry.listSteps())
        info = export_registry.getStepMetadata('one')
        self.assertEqual(info['id'], 'one')
        self.assertEqual(info['title'], 'One Step')
        self.assertIn('One small step', info['description'])
        self.assertEqual(info['handler'],
                         'Products.GenericSetup.tests.test_registry.ONE_FUNC')

        self.assertEqual(export_registry.getStep('one'), ONE_FUNC)

    def test_listContextInfos_empty(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        infos = tool.listContextInfos()
        self.assertEqual(len(infos), 0)

    def test_listContextInfos_with_snapshot(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        tool.createSnapshot('testing')
        infos = tool.listContextInfos()
        self.assertEqual(len(infos), 1)
        info = infos[0]
        self.assertEqual(info['id'], 'snapshot-testing')
        self.assertEqual(info['title'], 'testing')
        self.assertEqual(info['type'], 'snapshot')

    def test_listContextInfos_with_registered_base_profile(self):
        from ..interfaces import BASE

        profile_registry.registerProfile('foo', 'Foo', '', self._PROFILE_PATH,
                                         'Foo', BASE)
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        infos = tool.listContextInfos()
        self.assertEqual(len(infos), 1)
        info = infos[0]
        self.assertEqual(info['id'], 'profile-Foo:foo')
        self.assertEqual(info['title'], 'Foo')
        self.assertEqual(info['type'], 'base')

    def test_listContextInfos_with_registered_extension_profile(self):
        from ..interfaces import EXTENSION

        profile_registry.registerProfile('foo', 'Foo', '', self._PROFILE_PATH,
                                         'Foo', EXTENSION)
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        infos = tool.listContextInfos()
        self.assertEqual(len(infos), 1)
        info = infos[0]
        self.assertEqual(info['id'], 'profile-Foo:foo')
        self.assertEqual(info['title'], 'Foo')
        self.assertEqual(info['type'], 'extension')

    def test_listContextInfos_with_ordering(self):
        from ..interfaces import BASE
        from ..interfaces import EXTENSION

        # three extension profiles
        profile_registry.registerProfile('bar', 'bar', '', self._PROFILE_PATH,
                                         'bar', EXTENSION)
        profile_registry.registerProfile('foo', 'foo', '', self._PROFILE_PATH,
                                         'foo', EXTENSION)
        profile_registry.registerProfile('upper', 'UPPER', '',
                                         self._PROFILE_PATH, 'UPPER',
                                         EXTENSION)
        # one base profile
        profile_registry.registerProfile('base', 'base', '',
                                         self._PROFILE_PATH, 'base', BASE)

        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        tool.createSnapshot('UPPER')
        tool.createSnapshot('lower')
        infos = tool.listContextInfos()
        self.assertEqual(len(infos), 6)
        # We sort case insensitively, so by lowercase.
        # First snapshots.
        self.assertEqual(infos[0]['id'], 'snapshot-lower')
        self.assertEqual(infos[1]['id'], 'snapshot-UPPER')
        # Then base and extension profiles
        self.assertEqual(infos[2]['id'], 'profile-bar:bar')
        self.assertEqual(infos[3]['id'], 'profile-base:base')
        self.assertEqual(infos[4]['id'], 'profile-foo:foo')
        self.assertEqual(infos[5]['id'], 'profile-UPPER:upper')

    def test_getProfileImportDate_nonesuch(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        self.assertEqual(tool.getProfileImportDate('nonesuch'), None)

    def test_getProfileImportDate_simple_id(self):
        from OFS.Image import File

        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        filename = 'import-all-foo-20070315123456.log'
        tool._setObject(filename, File(filename, '', b''))
        self.assertEqual(tool.getProfileImportDate('foo'),
                         '2007-03-15T12:34:56Z')

    def test_getProfileImportDate_id_with_colon(self):
        from OFS.Image import File

        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        filename = 'import-all-foo_bar-20070315123456.log'
        tool._setObject(filename, File(filename, '', b''))
        self.assertEqual(tool.getProfileImportDate('foo:bar'),
                         '2007-03-15T12:34:56Z')

    def test_getProfileImportDate_id_with_prefix(self):
        # Test if getProfileImportDate does not fail if there is another
        # item id with id with a longer id which starts with the same
        # prefix
        from OFS.Image import File

        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        filename = 'import-all-foo_bar-20070315123456.log'
        tool._setObject(filename, File(filename, '', b''))
        filename2 = 'import-all-foo_bar-boo-20070315123456.log'
        tool._setObject(filename2, File(filename2, '', b''))
        self.assertEqual(tool.getProfileImportDate('foo:bar'),
                         '2007-03-15T12:34:56Z')

    def test_profileVersioning(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        profile_id = 'dummy_profile'
        product_name = 'GenericSetup'
        directory = os.path.split(__file__)[0]
        path = os.path.join(directory, 'versioned_profile')

        # register profile
        profile_registry.registerProfile(profile_id,
                                         'Dummy Profile',
                                         'This is a dummy profile',
                                         path,
                                         product=product_name)

        # register upgrade step
        step = UpgradeStep("Upgrade", "GenericSetup:dummy_profile", '*', '1.1',
                           '', dummy_upgrade, None, "1")
        _registerUpgradeStep(step)

        # test initial states
        profile_id = ':'.join((product_name, profile_id))
        self.assertEqual(tool.getVersionForProfile(profile_id), '1.1')
        self.assertEqual(tool.getLastVersionForProfile(profile_id), 'unknown')

        # run upgrade steps
        request = site.REQUEST
        request.form['profile_id'] = profile_id
        steps = listUpgradeSteps(tool, profile_id, '1.0')
        step_id = steps[0]['id']
        request.form['upgrades'] = [step_id]
        tool.manage_doUpgrades()
        self.assertEqual(tool.getLastVersionForProfile(profile_id), ('1', '1'))

    def test_get_and_setLastVersionForProfile(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        self.assertEqual(tool._profile_upgrade_versions, {})
        # Any 'profile-' is stripped off in these calls.
        self.assertEqual(tool.getLastVersionForProfile('foo'), 'unknown')
        self.assertEqual(tool.getLastVersionForProfile('profile-foo'),
                         'unknown')
        tool.setLastVersionForProfile('foo', '1.0')
        self.assertEqual(tool.getLastVersionForProfile('foo'), ('1', '0'))
        self.assertEqual(tool.getLastVersionForProfile('profile-foo'),
                         ('1', '0'))
        tool.setLastVersionForProfile('profile-foo', '2.0')
        self.assertEqual(tool.getLastVersionForProfile('foo'), ('2', '0'))
        self.assertEqual(tool.getLastVersionForProfile('profile-foo'),
                         ('2', '0'))

        # Setting the profile to unknown, removes it from the versions.
        self.assertEqual(tool._profile_upgrade_versions, {'foo': ('2', '0')})
        tool.setLastVersionForProfile('profile-foo', 'unknown')
        self.assertEqual(tool._profile_upgrade_versions, {})

    def test_unsetLastVersionForProfile(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        tool.setLastVersionForProfile('foo', '1.0')
        tool.setLastVersionForProfile('bar', '2.0')
        self.assertEqual(tool._profile_upgrade_versions, {
            'foo': ('1', '0'),
            'bar': ('2', '0')
        })

        # Any 'profile-' is stripped off in these calls.
        tool.unsetLastVersionForProfile('profile-foo')
        self.assertEqual(tool._profile_upgrade_versions, {'bar': ('2', '0')})
        tool.unsetLastVersionForProfile('bar')
        self.assertEqual(tool._profile_upgrade_versions, {})

    def test_purgeProfileVersions(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        tool.setLastVersionForProfile('foo', '1.0')
        tool.setLastVersionForProfile('bar', '2.0')
        self.assertEqual(tool._profile_upgrade_versions, {
            'foo': ('1', '0'),
            'bar': ('2', '0')
        })
        tool.purgeProfileVersions()
        self.assertEqual(tool._profile_upgrade_versions, {})

    def test_listUpgradeSteps(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool

        # we start without upgrades
        self.assertEqual(listUpgradeSteps(tool, 'prof', None), [])
        self.assertEqual(listUpgradeSteps(tool, 'prof', 'unknown'), [])
        self.assertEqual(listUpgradeSteps(tool, 'prof', '1.0'), [])
        self.assertEqual(listUpgradeSteps(tool, 'prof', '1.0', dest='1.1'), [])

        # register upgrade step
        step1 = UpgradeStep("Upgrade 1", 'prof', '1.0', '1.1', '',
                            dummy_upgrade, None, "1")
        _registerUpgradeStep(step1)
        self.assertEqual(len(listUpgradeSteps(tool, 'prof', 'unknown')), 1)
        self.assertEqual(len(listUpgradeSteps(tool, 'prof', None)), 1)
        self.assertEqual(len(listUpgradeSteps(tool, 'prof', '0.5')), 1)
        self.assertEqual(len(listUpgradeSteps(tool, 'prof', '1.0')), 1)
        self.assertEqual(len(listUpgradeSteps(tool, 'prof', '1.1')), 0)
        self.assertEqual(len(listUpgradeSteps(tool, 'prof', '1.2')), 0)
        # test with explicit destinations
        self.assertEqual(
            len(listUpgradeSteps(tool, 'prof', '0.5', dest='1.0')),
            0,
        )
        self.assertEqual(
            len(listUpgradeSteps(tool, 'prof', '0.5', dest='1.1')),
            1,
        )
        # 1.2 cannot be reached, but listUpgradeSteps does not care:
        # it reports all steps that bring the source in this direction.
        self.assertEqual(
            len(listUpgradeSteps(tool, 'prof', '0.5', dest='1.2')),
            1,
        )

        # register a second upgrade step
        step2 = UpgradeStep("Upgrade 2", 'prof', '1.1', '1.2', '',
                            dummy_upgrade, None, "2")
        _registerUpgradeStep(step2)
        self.assertEqual(len(listUpgradeSteps(tool, 'prof', 'unknown')), 2)
        self.assertEqual(len(listUpgradeSteps(tool, 'prof', None)), 2)
        self.assertEqual(len(listUpgradeSteps(tool, 'prof', '0.5')), 2)
        self.assertEqual(len(listUpgradeSteps(tool, 'prof', '1.0')), 2)
        self.assertEqual(len(listUpgradeSteps(tool, 'prof', '1.1')), 1)
        self.assertEqual(len(listUpgradeSteps(tool, 'prof', '1.2')), 0)
        self.assertEqual(
            len(listUpgradeSteps(tool, 'prof', '1.0', dest='1.0')),
            0,
        )
        self.assertEqual(
            len(listUpgradeSteps(tool, 'prof', '1.0', dest='1.1')),
            1,
        )
        self.assertEqual(
            len(listUpgradeSteps(tool, 'prof', '1.0', dest='1.2')),
            2,
        )
        self.assertEqual(
            len(listUpgradeSteps(tool, 'prof', '1.0', dest='1.3')),
            2,
        )
        # try different spelling of versions
        self.assertEqual(len(listUpgradeSteps(tool, 'prof', ('1', '1'))), 1)
        self.assertEqual(
            len(listUpgradeSteps(tool, 'prof', '1.0', dest=('1', '1'))),
            1,
        )
        self.assertEqual(
            len(listUpgradeSteps(tool, 'prof', ('1', '0'), dest=('1', '1'))),
            1,
        )

    def test_listUpgrades(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        profile_id_1 = 'dummy_profile1'
        product_name = 'GenericSetup'
        directory = os.path.split(__file__)[0]
        path = os.path.join(directory, 'versioned_profile')

        # register profile
        profile_registry.registerProfile(profile_id_1,
                                         'Dummy Profile 1',
                                         'This is dummy profile 1',
                                         path,
                                         product=product_name)

        # get full profile id
        profile_id_1 = ':'.join((product_name, profile_id_1))
        self.assertEqual(tool.listUpgrades(profile_id_1), [])
        self.assertEqual(tool.listUpgrades('non-existing'), [])
        self.assertFalse(tool.hasUpgrades(profile_id_1))
        self.assertFalse(tool.hasUpgrades('non-existing'))

        # register upgrade step
        step1 = UpgradeStep("Upgrade 1", profile_id_1, '1.0', '1.1', '',
                            dummy_upgrade, None, "1")
        _registerUpgradeStep(step1)
        self.assertTrue(tool.hasUpgrades(profile_id_1))
        self.assertFalse(tool.hasUpgrades('non-existing'))
        self.assertEqual(len(tool.listUpgrades(profile_id_1)), 1)
        self.assertEqual(len(tool.listUpgrades('non-existing')), 0)
        first_step = tool.listUpgrades(profile_id_1)[0]
        self.assertEqual(first_step.pop('id'), step1.id)
        self.assertIs(first_step.pop('step'), step1)
        self.assertEqual(
            first_step,
            {
                'description': '',
                'dest': ('1', '1'),
                'done': False,
                'haspath': ('1', '1'),
                'proposed': True,
                'sdest': '1.1',
                'sortkey': '1',
                'source': ('1', '0'),
                'ssource': '1.0',
                'title': 'Upgrade 1',
            },
        )
        # We can ask for a simple flat list, giving only steps.
        self.assertEqual(tool.listUpgrades(profile_id_1, simple=True), [step1])

        # register second upgrade step, with two sub steps
        # step2 = upgradeSteps(None, profile_id_1, '1.1', '1.2', '')
        step2_hash = 'step2_dummy_hash'
        step2a = UpgradeStep("Upgrade 2a", profile_id_1, '1.1', '1.2', '',
                             dummy_upgrade, None, "1")
        _registerNestedUpgradeStep(step2a, step2_hash)
        step2b = UpgradeDepends("Upgrade 2b",
                                profile_id_1,
                                '1.1',
                                '1.2',
                                '',
                                sortkey="2")
        _registerNestedUpgradeStep(step2b, step2_hash)
        self.assertTrue(tool.hasUpgrades(profile_id_1))
        self.assertEqual(len(tool.listUpgrades(profile_id_1)), 2)
        self.assertEqual(tool.listUpgrades(profile_id_1)[0]['id'], step1.id)
        self.assertEqual(
            tool.listUpgrades(profile_id_1)[1][0]['id'],
            step2a.id,
        )
        self.assertEqual(
            tool.listUpgrades(profile_id_1)[1][1]['id'],
            step2b.id,
        )
        self.assertEqual(len(tool.listUpgrades(profile_id_1, simple=True)), 3)
        self.assertEqual(
            tool.listUpgrades(profile_id_1, simple=True),
            [step1, step2a, step2b],
        )
        self.assertEqual(len(tool.listUpgrades(profile_id_1, dest='1.1')), 1)
        self.assertEqual(len(tool.listUpgrades(profile_id_1, dest='1.0')), 0)
        self.assertEqual(len(tool.listUpgrades(profile_id_1, dest='2.0')), 2)

        # Pretend profile 1 was installed to an earlier version.
        tool.setLastVersionForProfile(profile_id_1, '1.0')
        self.assertTrue(tool.hasUpgrades(profile_id_1))
        self.assertEqual(len(tool.listUpgrades(profile_id_1)), 2)
        self.assertEqual(len(tool.listUpgrades(profile_id_1, dest='1.1')), 1)
        self.assertEqual(len(tool.listUpgrades(profile_id_1, dest='1.0')), 0)
        self.assertEqual(len(tool.listUpgrades(profile_id_1, dest='2.0')), 2)
        self.assertEqual(
            tool.listUpgrades(profile_id_1, dest='1.1')[0]['id'],
            step1.id,
        )
        self.assertEqual(
            tool.listUpgrades(profile_id_1, simple=True),
            [step1, step2a, step2b],
        )

        # upgrade to 1.1
        tool.upgradeProfile(profile_id_1, dest='1.1')
        self.assertEqual(
            tool.getLastVersionForProfile(profile_id_1),
            ('1', '1'),
        )
        self.assertTrue(tool.hasUpgrades(profile_id_1))
        self.assertEqual(len(tool.listUpgrades(profile_id_1)), 1)
        self.assertEqual(
            len(tool.listUpgrades(profile_id_1, show_old=True)),
            2,
        )
        self.assertEqual(len(tool.listUpgrades(profile_id_1, dest='2.0')), 1)
        self.assertEqual(
            tool.listUpgrades(profile_id_1, simple=True),
            [step2a, step2b],
        )

        # upgrade all the way
        tool.upgradeProfile(profile_id_1)
        self.assertEqual(
            tool.getLastVersionForProfile(profile_id_1),
            ('1', '2'),
        )
        self.assertFalse(tool.hasUpgrades(profile_id_1))
        self.assertEqual(len(tool.listUpgrades(profile_id_1)), 0)
        self.assertEqual(
            len(tool.listUpgrades(profile_id_1, show_old=True)),
            2,
        )
        self.assertEqual(len(tool.listUpgrades(profile_id_1, dest='2.0')), 0)
        self.assertEqual(
            tool.listUpgrades(profile_id_1, simple=True),
            [],
        )
        self.assertEqual(
            tool.listUpgrades(profile_id_1, show_old=True, simple=True),
            [step1, step2a, step2b],
        )
        self.assertEqual(
            tool.listUpgrades(profile_id_1,
                              show_old=True,
                              simple=True,
                              dest='1.1'),
            [step1],
        )

    def test_listProfilesWithUpgrades(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        self.assertEqual(tool.listProfilesWithUpgrades(), [])
        self.assertEqual(tool.listProfilesWithPendingUpgrades(), [])
        self.assertEqual(tool.listUptodateProfiles(), [])
        self.assertEqual(tool.hasPendingUpgrades(), False)
        profile_id = 'dummy_profile'
        product_name = 'GenericSetup'
        directory = os.path.split(__file__)[0]
        path = os.path.join(directory, 'versioned_profile')

        # register profile
        profile_registry.registerProfile(profile_id,
                                         'Dummy Profile',
                                         'This is a dummy profile',
                                         path,
                                         product=product_name)
        self.assertEqual(tool.listProfilesWithUpgrades(), [])
        self.assertEqual(tool.listProfilesWithPendingUpgrades(), [])
        self.assertEqual(tool.listUptodateProfiles(), [])
        self.assertEqual(tool.hasPendingUpgrades(), False)

        # register upgrade step
        step1 = UpgradeStep("Upgrade 1", "GenericSetup:dummy_profile", '*',
                            '1.1', '', dummy_upgrade, None, "1")
        _registerUpgradeStep(step1)
        self.assertEqual(tool.listProfilesWithUpgrades(),
                         ['GenericSetup:dummy_profile'])
        self.assertEqual(tool.listProfilesWithPendingUpgrades(), [])
        self.assertEqual(tool.listUptodateProfiles(), [])
        self.assertEqual(tool.hasPendingUpgrades(), False)

        # register another upgrade step
        step2 = UpgradeStep("Upgrade 2", "GenericSetup:dummy_profile", '1.1',
                            '1.2', '', dummy_upgrade, None, "1")
        _registerUpgradeStep(step2)
        self.assertEqual(tool.listProfilesWithUpgrades(),
                         ['GenericSetup:dummy_profile'])
        self.assertEqual(tool.listProfilesWithPendingUpgrades(), [])
        self.assertEqual(tool.listUptodateProfiles(), [])
        self.assertEqual(tool.hasPendingUpgrades(), False)

        # get full profile id
        profile_id = ':'.join((product_name, profile_id))

        # Pretend the profile was installed
        tool.setLastVersionForProfile(profile_id, '1.0')
        self.assertEqual(tool.listProfilesWithUpgrades(),
                         ['GenericSetup:dummy_profile'])
        self.assertEqual(tool.listProfilesWithPendingUpgrades(),
                         ['GenericSetup:dummy_profile'])
        self.assertEqual(tool.listUptodateProfiles(), [])
        self.assertEqual(tool.hasPendingUpgrades(), True)

        # run first upgrade step
        request = site.REQUEST
        request.form['profile_id'] = profile_id
        steps = listUpgradeSteps(tool, profile_id, '1.0')
        step_id = steps[0]['id']
        request.form['upgrades'] = [step_id]
        tool.manage_doUpgrades()
        self.assertEqual(tool.getLastVersionForProfile(profile_id), ('1', '1'))
        self.assertEqual(tool.listProfilesWithUpgrades(),
                         ['GenericSetup:dummy_profile'])
        self.assertEqual(tool.listProfilesWithPendingUpgrades(),
                         ['GenericSetup:dummy_profile'])
        self.assertEqual(tool.listUptodateProfiles(), [])
        self.assertEqual(tool.hasPendingUpgrades(), True)

        # run second upgrade step
        request = site.REQUEST
        request.form['profile_id'] = profile_id
        steps = listUpgradeSteps(tool, profile_id, '1.1')
        step_id = steps[0]['id']
        request.form['upgrades'] = [step_id]
        tool.manage_doUpgrades()
        self.assertEqual(tool.getLastVersionForProfile(profile_id), ('1', '2'))
        self.assertEqual(tool.listProfilesWithUpgrades(),
                         ['GenericSetup:dummy_profile'])
        self.assertEqual(tool.listProfilesWithPendingUpgrades(), [])
        self.assertEqual(tool.listUptodateProfiles(),
                         ['GenericSetup:dummy_profile'])
        self.assertEqual(tool.hasPendingUpgrades(), False)

        # Pretend the profile was never installed.
        tool.unsetLastVersionForProfile(profile_id)
        self.assertEqual(tool.listProfilesWithPendingUpgrades(), [])
        self.assertEqual(tool.listUptodateProfiles(), [])
        self.assertEqual(tool.hasPendingUpgrades(), False)

    def test_hasPendingUpgrades(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        profile_id_1 = 'dummy_profile1'
        profile_id_2 = 'dummy_profile2'
        product_name = 'GenericSetup'
        directory = os.path.split(__file__)[0]
        path = os.path.join(directory, 'versioned_profile')

        # register profiles
        profile_registry.registerProfile(profile_id_1,
                                         'Dummy Profile 1',
                                         'This is dummy profile 1',
                                         path,
                                         product=product_name)
        profile_registry.registerProfile(profile_id_2,
                                         'Dummy Profile 2',
                                         'This is dummy profile 2',
                                         path,
                                         product=product_name)

        # get full profile ids
        profile_id_1 = ':'.join((product_name, profile_id_1))
        profile_id_2 = ':'.join((product_name, profile_id_2))

        # test
        self.assertEqual(tool.hasPendingUpgrades(), False)
        self.assertEqual(tool.hasPendingUpgrades(profile_id_1), False)
        self.assertEqual(tool.hasPendingUpgrades(profile_id_2), False)
        self.assertEqual(tool.hasPendingUpgrades('non-existing'), False)

        # register upgrade steps
        step1 = UpgradeStep("Upgrade 1", profile_id_1, '*', '1.1', '',
                            dummy_upgrade, None, "1")
        _registerUpgradeStep(step1)
        step2 = UpgradeStep("Upgrade 2", profile_id_2, '*', '2.2', '',
                            dummy_upgrade, None, "2")
        _registerUpgradeStep(step2)
        # No profile has been applied, so no upgrade is pending.
        self.assertEqual(tool.hasPendingUpgrades(), False)
        self.assertEqual(tool.hasPendingUpgrades(profile_id_1), False)
        self.assertEqual(tool.hasPendingUpgrades(profile_id_2), False)

        # Pretend profile 1 was installed to an earlier version.
        tool.setLastVersionForProfile(profile_id_1, '1.0')
        self.assertEqual(tool.hasPendingUpgrades(), True)
        self.assertEqual(tool.hasPendingUpgrades(profile_id_1), True)
        self.assertEqual(tool.hasPendingUpgrades(profile_id_2), False)

        # Pretend profile 2 was installed to an earlier version.
        tool.setLastVersionForProfile(profile_id_2, '2.0')
        self.assertEqual(tool.hasPendingUpgrades(), True)
        self.assertEqual(tool.hasPendingUpgrades(profile_id_1), True)
        self.assertEqual(tool.hasPendingUpgrades(profile_id_2), True)

        # Pretend profile 1 was installed to the final version.
        tool.setLastVersionForProfile(profile_id_1, '1.1')
        self.assertEqual(tool.hasPendingUpgrades(), True)
        self.assertEqual(tool.hasPendingUpgrades(profile_id_1), False)
        self.assertEqual(tool.hasPendingUpgrades(profile_id_2), True)

        # Pretend profile 2 was installed to the final version.
        tool.setLastVersionForProfile(profile_id_2, '2.2')
        self.assertEqual(tool.hasPendingUpgrades(), False)
        self.assertEqual(tool.hasPendingUpgrades(profile_id_1), False)
        self.assertEqual(tool.hasPendingUpgrades(profile_id_2), False)

    def test_manage_doUpgrades_no_profile_id_or_updates(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        tool.manage_doUpgrades()
        self.assertEqual(tool._profile_upgrade_versions, {})

    def test_manage_doUpgrades_upgrade_w_no_target_version(self):

        def notool():
            return None

        step = UpgradeStep('TITLE', 'foo', '*', '*', 'DESC', notool)
        _registerUpgradeStep(step)
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        request = site.REQUEST
        request['profile_id'] = ['foo']
        request['upgrade'] = [step.id]
        tool.manage_doUpgrades()
        self.assertEqual(tool._profile_upgrade_versions, {})

    def test_upgradeProfile_no_profile_id_or_updates(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        # Mostly this checks to see if we can call this without an
        # exception.
        tool.upgradeProfile('no.such.profile:default')
        self.assertEqual(tool._profile_upgrade_versions, {})
        tool.upgradeProfile('no.such.profile:default', dest='42')
        self.assertEqual(tool._profile_upgrade_versions, {})

        # XXX We could use self.assertLogs to check that a warning message was
        # logged, and on 3.10+ use self.assertNoLogs to test the opposite when
        # called with quiet=True.  For now quickly test that the quiet argument
        # throws no error.
        tool.upgradeProfile('no.such.profile:default', quiet=True)

    def test_persistent_profile_upgrade_versions(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        savepoint1 = transaction.savepoint()
        tool.setLastVersionForProfile('foo', '1.0')
        savepoint2 = transaction.savepoint()
        tool.setLastVersionForProfile('bar', '2.0')
        self.assertEqual(tool._profile_upgrade_versions, {
            'foo': ('1', '0'),
            'bar': ('2', '0')
        })
        savepoint2.rollback()
        self.assertEqual(tool._profile_upgrade_versions, {'foo': ('1', '0')})
        savepoint1.rollback()
        self.assertEqual(tool._profile_upgrade_versions, {})

    def test_separate_profile_upgrade_versions(self):
        # _profile_upgrade_versions used to be a class property.  That is fine
        # as long as we only work on copies, otherwise state is shared between
        # two instances.  We now create the property in the __init__ method,
        # but let's test it to avoid a regression.
        site = self._makeSite()
        site.setup_tool1 = self._makeOne('setup_tool1')
        tool1 = site.setup_tool1
        site.setup_tool2 = self._makeOne('setup_tool2')
        tool2 = site.setup_tool2
        tool1._profile_upgrade_versions['foo'] = '1.0'
        self.assertEqual(tool2._profile_upgrade_versions, {})
        tool2.setLastVersionForProfile('bar', '2.0')
        self.assertEqual(self._makeOne('t')._profile_upgrade_versions, {})

    def test_upgradeProfile(self):

        def dummy_handler(tool):
            return None

        def step3_handler(tool):
            tool._step3_applied = 'just a marker'

        def step3_checker(tool):
            # False means already applied or does not apply.
            # True means can be applied.
            return not hasattr(tool, '_step3_applied')

        step1 = UpgradeStep('Step 1', 'foo', '0', '1', 'DESC', dummy_handler)
        step2 = UpgradeStep('Step 2', 'foo', '1', '2', 'DESC', dummy_handler)
        step3 = UpgradeStep('Step 3',
                            'foo',
                            '2',
                            '3',
                            'DESC',
                            step3_handler,
                            checker=step3_checker)
        step4 = UpgradeStep('Step 4', 'foo', '3', '4', 'DESC', dummy_handler)
        _registerUpgradeStep(step1)
        _registerUpgradeStep(step2)
        _registerUpgradeStep(step3)
        _registerUpgradeStep(step4)
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        self.assertEqual(tool.getLastVersionForProfile('foo'), 'unknown')
        tool.setLastVersionForProfile('foo', '0')
        self.assertEqual(tool.getLastVersionForProfile('foo'), ('0', ))
        # Upgrade the profile one step to version 1.
        tool.upgradeProfile('foo', '1')
        self.assertEqual(tool.getLastVersionForProfile('foo'), ('1', ))
        # Do that again and you will get a warning, which you can choose
        # to silence.  Note: we do not test if something really gets logged.
        tool.upgradeProfile('foo', '1')
        self.assertEqual(tool.getLastVersionForProfile('foo'), ('1', ))
        tool.upgradeProfile('foo', '1', quiet=True)
        self.assertEqual(tool.getLastVersionForProfile('foo'), ('1', ))
        # Upgrade the profile two steps to version 3.  This one has a
        # checker.  The profile version must be correctly updated.
        tool.upgradeProfile('foo', '3')
        self.assertEqual(tool.getLastVersionForProfile('foo'), ('3', ))
        # Upgrade the profile to a non existing version.  Nothing
        # should happen.
        tool.upgradeProfile('foo', '5')
        self.assertEqual(tool.getLastVersionForProfile('foo'), ('3', ))
        # It throws a warning which you can choose to ignore.
        # The effect is the same.
        tool.upgradeProfile('foo', '5', quiet=True)
        self.assertEqual(tool.getLastVersionForProfile('foo'), ('3', ))
        # Upgrade the profile to the latest version.
        tool.upgradeProfile('foo')
        self.assertEqual(tool.getLastVersionForProfile('foo'), ('4', ))

    def test_listExportSteps(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        result = tool.listExportSteps()
        self.assertEqual(len(result), 4)
        self.assertIn('componentregistry', result)
        self.assertIn('rolemap', result)
        self.assertIn('step_registries', result)
        self.assertIn('toolset', result)

        tool._export_registry.registerStep('foo', handler='foo.export')
        tool._export_registry.registerStep('toolset', handler='toolset.export')
        result = tool.listExportSteps()
        self.assertEqual(len(result), 5)
        self.assertIn('componentregistry', result)
        self.assertIn('foo', result)
        self.assertIn('rolemap', result)
        self.assertIn('step_registries', result)
        self.assertIn('toolset', result)

    def test_getSortedImportSteps(self):
        site = self._makeSite()
        site.setup_tool = self._makeOne('setup_tool')
        tool = site.setup_tool
        result = tool.getSortedImportSteps()
        self.assertEqual(len(result), 3)
        self.assertIn('componentregistry', result)
        self.assertIn('rolemap', result)
        self.assertIn('toolset', result)
        self.assertGreater(
            list(result).index('componentregistry'),
            list(result).index(
                'toolset')
        )

        tool._import_registry.registerStep('foo', handler='foo.import')
        tool._import_registry.registerStep('toolset', handler='toolset.import')
        result = tool.getSortedImportSteps()
        self.assertEqual(len(result), 4)
        self.assertIn('componentregistry', result)
        self.assertIn('foo', result)
        self.assertIn('rolemap', result)
        self.assertIn('toolset', result)
        self.assertGreater(
            list(result).index('componentregistry'),
            list(result).index(
                'toolset')
        )

    def test_listProfileInfo_for_parameter(self):
        from ..metadata import METADATA_XML

        self._makeFile(METADATA_XML, _METADATA_XML)

        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)

        profile_registry.registerProfile('foo',
                                         'Foo',
                                         '',
                                         self._PROFILE_PATH,
                                         for_=ISite)
        # tool.listProfileInfo should call registry.listProfileInfo
        # with the for_ parameter
        self.assertEqual(len(tool.listProfileInfo()), 1)
        self.assertEqual(len(tool.listProfileInfo(for_=ISite)), 1)
        self.assertEqual(len(tool.listProfileInfo(for_=IDerivedSite)), 1)
        self.assertEqual(len(tool.listProfileInfo(for_=IAnotherSite)), 0)

    def test_profileExists(self):
        from ..interfaces import EXTENSION
        from ..metadata import METADATA_XML
        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)

        # Register two extension profiles.  Profile 'foo' has a dependency
        # 'bar'.
        self._makeFile(METADATA_XML, _METADATA_XML)
        _makeTestFile(METADATA_XML, self._PROFILE_PATH2, _PLAIN_METADATA_XML)
        profile_registry.registerProfile('foo',
                                         'Foo',
                                         '',
                                         self._PROFILE_PATH,
                                         profile_type=EXTENSION)
        profile_registry.registerProfile('bar',
                                         'Bar',
                                         '',
                                         self._PROFILE_PATH2,
                                         profile_type=EXTENSION)

        self.assertTrue(tool.profileExists('other:foo'))
        self.assertTrue(tool.profileExists('other:bar'))
        self.assertFalse(tool.profileExists('snapshot-something'))
        self.assertFalse(tool.profileExists(None))
        self.assertFalse(tool.profileExists('nonesuch'))

    def test_getDependenciesForProfile(self):
        from ..interfaces import EXTENSION
        from ..metadata import METADATA_XML
        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)

        # Register three extension profiles.  Profile 'foo' has a dependency
        # 'bar', and 'baz' contains non-existing dependency-profiles.
        self._makeFile(METADATA_XML, _METADATA_XML)
        _makeTestFile(METADATA_XML, self._PROFILE_PATH2, _PLAIN_METADATA_XML)
        _makeTestFile(METADATA_XML, self._PROFILE_PATH3, _BROKEN_METADATA_XML)
        profile_registry.registerProfile('foo',
                                         'Foo',
                                         '',
                                         self._PROFILE_PATH,
                                         profile_type=EXTENSION)
        profile_registry.registerProfile('bar',
                                         'Bar',
                                         '',
                                         self._PROFILE_PATH2,
                                         profile_type=EXTENSION)
        profile_registry.registerProfile('baz',
                                         'Baz',
                                         '',
                                         self._PROFILE_PATH3,
                                         profile_type=EXTENSION)

        self.assertEqual(tool.getDependenciesForProfile('other:foo'),
                         ('profile-other:bar', ))
        self.assertEqual(tool.getDependenciesForProfile('other:bar'), ())
        self.assertEqual(tool.getDependenciesForProfile('snapshot-some'), ())
        self.assertEqual(tool.getDependenciesForProfile(None), ())
        self.assertRaises(KeyError, tool.getDependenciesForProfile, 'nonesuch')

    def test_getBrokenDependencies(self):
        from ..interfaces import EXTENSION
        from ..metadata import METADATA_XML
        site = self._makeSite()
        tool = self._makeOne('setup_tool').__of__(site)

        # Register three extension profiles.  Profile 'foo' has a dependency
        # 'bar' and 'baz' contains non-existing dependency-profiles.
        self._makeFile(METADATA_XML, _METADATA_XML)
        _makeTestFile(METADATA_XML, self._PROFILE_PATH2, _PLAIN_METADATA_XML)
        _makeTestFile(METADATA_XML, self._PROFILE_PATH3, _BROKEN_METADATA_XML)
        profile_registry.registerProfile('foo',
                                         'Foo',
                                         '',
                                         self._PROFILE_PATH,
                                         profile_type=EXTENSION)
        profile_registry.registerProfile('bar',
                                         'Bar',
                                         '',
                                         self._PROFILE_PATH2,
                                         profile_type=EXTENSION)
        profile_registry.registerProfile('baz',
                                         'Baz',
                                         '',
                                         self._PROFILE_PATH3,
                                         profile_type=EXTENSION)

        # profile has dependencies and none of them is broken:
        self.assertFalse(tool.hasBrokenDependencies('other:foo'))
        # profile has no dependencies, therfore nothing can be broken:
        self.assertFalse(tool.hasBrokenDependencies('other:bar'))
        # profile has dependencies and at least one of them is broken:
        self.assertTrue(tool.hasBrokenDependencies('other:baz'))


_DEFAULT_STEP_REGISTRIES_EXPORT_XML = ("""\
<?xml version="1.0"?>
<export-steps>
 <export-step id="step_registries"
              handler="Products.GenericSetup.tool.exportStepRegistries"
              title="Export import / export steps.">
""" + "  " + """
 </export-step>
</export-steps>
""")

_EXTRAS_STEP_REGISTRIES_EXPORT_XML = """\
<?xml version="1.0"?>
<export-steps>
 <export-step
    id="properties"
    handler="Products.GenericSetup.tests.test_tool._exportPropertiesINI"
    title="properties">

 </export-step>
 <export-step
    id="step_registries"
    handler="Products.GenericSetup.tool.exportStepRegistries"
    title="Export import / export steps.">

 </export-step>
</export-steps>
"""

_DEFAULT_STEP_REGISTRIES_IMPORT_XML = ("""\
<?xml version="1.0"?>
<import-steps>
 <import-step id="foo" handler="foo.bar" title="foo">
""" + "  " + """
 </import-step>
</import-steps>
""")

_EXTRAS_STEP_REGISTRIES_IMPORT_XML = """\
<?xml version="1.0"?>
<import-steps>
 <import-step
    id="dependable"
    version="1"
    handler="Products.GenericSetup.tests.test_tool._underscoreSiteTitle"
    title="dependable">
  <dependency step="purging" />

 </import-step>
 <import-step
    id="dependent"
    version="1"
    handler="Products.GenericSetup.tests.test_tool._uppercaseSiteTitle"
    title="dependent">
  <dependency step="dependable" />

 </import-step>
 <import-step
    id="purging"
    version="1"
    handler="Products.GenericSetup.tests.test_tool._purgeIfRequired"
    title="purging">

 </import-step>
</import-steps>
"""

_PROPERTIES_INI = """\
[Default]
Title=%s
"""


def _underscoreSiteTitle(context):

    site = context.getSite()
    site.title = site.title.replace(' ', '_')
    return 'Underscored title'


def _uppercaseSiteTitle(context):

    site = context.getSite()
    site.title = site.title.upper()
    return 'Uppercased title'


def _purgeIfRequired(context):

    site = context.getSite()
    purged = site.purged = context.shouldPurge()
    return purged and 'Purged' or 'Unpurged'


def _exportPropertiesINI(context):

    site = context.getSite()
    text = _PROPERTIES_INI % site.title

    context.writeDataFile('properties.ini', text.encode('utf-8'), 'text/plain')

    return 'Exported properties'


class _ToolsetSetup(BaseRegistryTests):

    def _initSite(self):
        from ..tool import SetupTool

        site = Folder()
        site._setId('site')
        self.app._setObject('site', site)
        site = self.app._getOb('site')
        site._setObject('setup_tool', SetupTool('setup_tool'))
        return site


class Test_exportToolset(_ToolsetSetup):

    layer = ExportImportZCMLLayer

    def test_empty(self):
        from ..tool import TOOLSET_XML
        from ..tool import exportToolset

        site = self._initSite()
        context = DummyExportContext(site, tool=site.setup_tool)

        exportToolset(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, TOOLSET_XML)
        self._compareDOM(text.decode('utf-8'), _EMPTY_TOOLSET_XML)
        self.assertEqual(content_type, 'text/xml')

    def test_normal(self):
        from ..tool import TOOLSET_XML
        from ..tool import exportToolset

        site = self._initSite()
        toolset = site.setup_tool.getToolsetRegistry()
        toolset.addForbiddenTool('doomed')
        toolset.addRequiredTool('mandatory', 'path.to.one')
        toolset.addRequiredTool('obligatory', 'path.to.another')

        context = DummyExportContext(site, tool=site.setup_tool)

        exportToolset(context)

        self.assertEqual(len(context._wrote), 1)
        filename, text, content_type = context._wrote[0]
        self.assertEqual(filename, TOOLSET_XML)
        self._compareDOM(text.decode('utf-8'), _NORMAL_TOOLSET_XML)
        self.assertEqual(content_type, 'text/xml')


class Test_importToolset(_ToolsetSetup):

    layer = ExportImportZCMLLayer

    def test_import_updates_registry(self):
        from ..tool import TOOLSET_XML
        from ..tool import importToolset

        site = self._initSite()
        context = DummyImportContext(site, tool=site.setup_tool)

        # Import forbidden
        context._files[TOOLSET_XML] = _FORBIDDEN_TOOLSET_XML
        importToolset(context)

        tool = context.getSetupTool()
        toolset = tool.getToolsetRegistry()

        self.assertEqual(len(toolset.listForbiddenTools()), 3)
        self.assertIn('doomed', toolset.listForbiddenTools())
        self.assertIn('damned', toolset.listForbiddenTools())
        self.assertIn('blasted', toolset.listForbiddenTools())

        # Import required
        context._files[TOOLSET_XML] = _REQUIRED_TOOLSET_XML
        importToolset(context)

        self.assertEqual(len(toolset.listRequiredTools()), 2)
        self.assertIn('mandatory', toolset.listRequiredTools())
        info = toolset.getRequiredToolInfo('mandatory')
        self.assertEqual(info['class'],
                         'Products.GenericSetup.tests.test_tool.DummyTool')
        self.assertIn('obligatory', toolset.listRequiredTools())
        info = toolset.getRequiredToolInfo('obligatory')
        self.assertEqual(info['class'],
                         'Products.GenericSetup.tests.test_tool.DummyTool')

    def test_tool_ids(self):
        # The tool import mechanism used to rely on the fact that all tools
        # have unique IDs set at the class level and that you can call their
        # constructor with no arguments. However, there might be tools
        # that need IDs set.
        from ..tool import TOOLSET_XML
        from ..tool import importToolset

        site = self._initSite()
        context = DummyImportContext(site, tool=site.setup_tool)
        context._files[TOOLSET_XML] = _REQUIRED_TOOLSET_XML

        importToolset(context)

        for tool_id in ('mandatory', 'obligatory'):
            tool = getattr(site, tool_id)
            self.assertEqual(tool.getId(), tool_id)

    def test_tool_id_required(self):
        # Tests that tool creation will still work when an id is required
        # by the tool constructor.
        from ..tool import TOOLSET_XML
        from ..tool import importToolset

        site = self._initSite()
        context = DummyImportContext(site, tool=site.setup_tool)
        context._files[TOOLSET_XML] = _WITH_ID_TOOLSET_XML

        importToolset(context)

        for tool_id in ('mandatory', 'requires_id', 'immutable_id'):
            tool = getattr(site, tool_id)
            self.assertEqual(tool.getId(), tool_id)

    def test_forbidden_tools(self):
        from ..tool import TOOLSET_XML
        from ..tool import importToolset

        TOOL_IDS = ('doomed', 'blasted', 'saved')

        site = self._initSite()

        for tool_id in TOOL_IDS:
            pseudo = Folder()
            pseudo._setId(tool_id)
            site._setObject(tool_id, pseudo)

        self.assertEqual(len(site.objectIds()), len(TOOL_IDS) + 1)

        for tool_id in TOOL_IDS:
            self.assertIn(tool_id, site.objectIds())

        context = DummyImportContext(site, tool=site.setup_tool)
        context._files[TOOLSET_XML] = _FORBIDDEN_TOOLSET_XML

        importToolset(context)

        self.assertEqual(len(site.objectIds()), 2)
        self.assertIn('setup_tool', site.objectIds())
        self.assertIn('saved', site.objectIds())

    def test_required_tools_missing(self):
        from ..tool import TOOLSET_XML
        from ..tool import importToolset

        site = self._initSite()
        self.assertEqual(len(site.objectIds()), 1)

        context = DummyImportContext(site, tool=site.setup_tool)
        context._files[TOOLSET_XML] = _REQUIRED_TOOLSET_XML

        importToolset(context)

        self.assertEqual(len(site.objectIds()), 3)
        self.assertIsInstance(
            aq_base(site._getOb('mandatory')),
            DummyTool
        )
        self.assertIsInstance(
            aq_base(site._getOb('obligatory')),
            DummyTool
        )

    def test_required_tools_no_replacement(self):
        from ..tool import TOOLSET_XML
        from ..tool import importToolset

        site = self._initSite()

        mandatory = DummyTool()
        mandatory._setId('mandatory')
        site._setObject('mandatory', mandatory)

        obligatory = DummyTool()
        obligatory._setId('obligatory')
        site._setObject('obligatory', obligatory)

        self.assertEqual(len(site.objectIds()), 3)

        context = DummyImportContext(site, tool=site.setup_tool)
        context._files[TOOLSET_XML] = _REQUIRED_TOOLSET_XML

        importToolset(context)

        self.assertEqual(len(site.objectIds()), 3)
        self.assertIs(aq_base(site._getOb('mandatory')), mandatory)
        self.assertIs(aq_base(site._getOb('obligatory')), obligatory)

    def test_required_tools_with_replacement(self):
        from ..tool import TOOLSET_XML
        from ..tool import importToolset

        site = self._initSite()

        mandatory = AnotherDummyTool()
        mandatory._setId('mandatory')
        site._setObject('mandatory', mandatory)

        obligatory = SubclassedDummyTool()
        obligatory._setId('obligatory')
        site._setObject('obligatory', obligatory)

        self.assertEqual(len(site.objectIds()), 3)

        context = DummyImportContext(site, tool=site.setup_tool)
        context._files[TOOLSET_XML] = _REQUIRED_TOOLSET_XML

        importToolset(context)

        self.assertEqual(len(site.objectIds()), 3)

        self.assertIsNot(aq_base(site._getOb('mandatory')), mandatory)
        self.assertIsInstance(
            aq_base(site._getOb('mandatory')),
            DummyTool
        )

        self.assertIsNot(aq_base(site._getOb('obligatory')), obligatory)
        self.assertIsInstance(
            aq_base(site._getOb('obligatory')),
            DummyTool
        )

    def test_required_tools_missing_acquired_nofail(self):
        from ..tool import TOOLSET_XML
        from ..tool import importToolset

        site = self._initSite()
        parent_site = Folder()

        mandatory = AnotherDummyTool()
        mandatory._setId('mandatory')
        parent_site._setObject('mandatory', mandatory)

        obligatory = AnotherDummyTool()
        obligatory._setId('obligatory')
        parent_site._setObject('obligatory', obligatory)

        site = site.__of__(parent_site)

        # acquiring subobjects of a different class during import
        # should not prevent new objects from being created if they
        # don't exist in the site

        context = DummyImportContext(site, tool=site.setup_tool)
        context._files[TOOLSET_XML] = _REQUIRED_TOOLSET_XML

        importToolset(context)

        self.assertIsNot(aq_base(site._getOb('mandatory')), mandatory)
        self.assertIsInstance(
            aq_base(site._getOb('mandatory')),
            DummyTool
        )

        self.assertIsNot(aq_base(site._getOb('obligatory')), obligatory)
        self.assertIsInstance(
            aq_base(site._getOb('obligatory')),
            DummyTool
        )

    def test_required_tools_missing_class_with_replacement(self):
        from ..tool import TOOLSET_XML
        from ..tool import importToolset

        site = self._initSite()

        obligatory = AnotherDummyTool()
        obligatory._setId('obligatory')
        site._setObject('obligatory', obligatory)

        self.assertEqual(len(site.objectIds()), 2)

        context = DummyImportContext(site, tool=site.setup_tool)
        context._files[TOOLSET_XML] = _BAD_CLASS_TOOLSET_XML

        importToolset(context)

        self.assertEqual(len(site.objectIds()), 2)


class DummyTool(Folder):

    pass


class AnotherDummyTool(Folder):

    pass


class SubclassedDummyTool(DummyTool):

    pass


class DummyToolRequiresId(Folder):

    def __init__(self, id):
        Folder.__init__(self)
        self._setId(id)


class DummyToolImmutableId(Folder):

    id = 'immutable_id'

    def _setId(self, id):
        if id != self.getId():
            raise ValueError()


def pre_handler(tool):
    try:
        tool.pre_handler_called += 1
    except AttributeError:
        tool.pre_handler_called = 1


def post_handler(tool):
    try:
        tool.post_handler_called += 1
    except AttributeError:
        tool.post_handler_called = 1


_FORBIDDEN_TOOLSET_XML = """\
<?xml version="1.0"?>
<tool-setup>
 <forbidden tool_id="doomed" />
 <forbidden tool_id="damned" />
 <forbidden tool_id="blasted" />
</tool-setup>
"""

_REQUIRED_TOOLSET_XML = """\
<?xml version="1.0"?>
<tool-setup>
 <required
    tool_id="mandatory"
    class="Products.GenericSetup.tests.test_tool.DummyTool" />
 <required
    tool_id="obligatory"
    class="Products.GenericSetup.tests.test_tool.DummyTool" />
</tool-setup>
"""

_WITH_ID_TOOLSET_XML = """\
<?xml version="1.0"?>
<tool-setup>
  <required
    tool_id="mandatory"
    class="Products.GenericSetup.tests.test_tool.DummyTool" />
  <required
    tool_id="requires_id"
    class="Products.GenericSetup.tests.test_tool.DummyToolRequiresId" />
  <required
    tool_id="immutable_id"
    class="Products.GenericSetup.tests.test_tool.DummyToolImmutableId" />
</tool-setup>
"""

_BAD_CLASS_TOOLSET_XML = """\
<?xml version="1.0"?>
<tool-setup>
 <required
    tool_id="obligatory"
    class="foobar" />
</tool-setup>
"""


def test_suite():
    return unittest.TestSuite((
        unittest.defaultTestLoader.loadTestsFromTestCase(SetupToolTests),
        unittest.defaultTestLoader.loadTestsFromTestCase(Test_exportToolset),
        unittest.defaultTestLoader.loadTestsFromTestCase(Test_importToolset),
    ))



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Components setup view.
"""

from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.Five.component.interfaces import IObjectManagerSite
from zope.component import adapts
from zope.component import getAdapter
from zope.component import getMultiAdapter
from zope.interface import Interface
from zope.interface import implementer
from zope.schema import Text

from ..context import SetupEnviron
from ..interfaces import IBody


class IComponentsSetupSchema(Interface):

    """Schema for components setup views.
    """

    body = Text(
        title='Settings')


@implementer(IComponentsSetupSchema)
class ComponentsSetupSchemaAdapter:

    adapts(IObjectManagerSite)

    def __init__(self, context):
        self.context = context

    def _getBody(self):
        sm = self.context.aq_inner.getSiteManager()
        return getMultiAdapter((sm, SetupEnviron()), IBody).body

    def _setBody(self, value):
        sm = self.context.aq_inner.getSiteManager()
        getMultiAdapter((sm, SetupEnviron()), IBody).body = value

    body = property(_getBody, _setBody)


class ComponentsSetupView(BrowserView):

    """Components setup view for IObjectManagerSite.
    """

    template = form_template = ViewPageTemplateFile('components_form.pt')
    status = ''

    def update(self):
        self.adapter = getAdapter(self.context, IComponentsSetupSchema)

        if 'apply' in self.request.form:
            self.adapter.body = self.request.form['body']
            self.status = 'Saved changes.'

    def __call__(self):
        self.update()
        return self.template()


class ComponentsSetupTab(ComponentsSetupView):

    """Components setup ZMI tab for IObjectManagerSite.
    """

    template = ViewPageTemplateFile('components.pt')



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Unit tests for import/export step zcml module.
"""

import unittest

from Zope2.App import zcml
from zope.testing.cleanup import cleanUp

import Products.GenericSetup

from ..registry import _import_step_registry


EMPTY_ZCML = '''\
<configure xmlns:genericsetup="http://namespaces.zope.org/genericsetup">
</configure>'''

ONE_STEP_ZCML = '''\
<configure xmlns:genericsetup="http://namespaces.zope.org/genericsetup"
           i18n_domain="genericsetup">
<genericsetup:importStep
    name="Products.GenericSetup.teststep"
    title="step title"
    description="step description"
    handler="Products.GenericSetup.initialize"
    />
</configure>'''


class ImportStepTests(unittest.TestCase):

    def setUp(self):
        zcml.load_config('meta.zcml', Products.GenericSetup)

    def tearDown(self):
        cleanUp()

    def testEmptyImport(self):
        zcml.load_string(EMPTY_ZCML)
        self.assertEqual(len(_import_step_registry.listSteps()), 0)

    def testOneStepImport(self):
        zcml.load_string(ONE_STEP_ZCML)
        self.assertEqual(_import_step_registry.listSteps(),
                         ['Products.GenericSetup.teststep'])
        info = _import_step_registry.getStepMetadata(
            'Products.GenericSetup.teststep')
        self.assertEqual(info['description'],
                         'step description')
        self.assertEqual(info['title'],
                         'step title')
        self.assertEqual(info['handler'],
                         'Products.GenericSetup.initialize')
        self.assertEqual(info['id'],
                         'Products.GenericSetup.teststep')


def test_suite():
    return unittest.TestSuite((
        unittest.defaultTestLoader.loadTestsFromTestCase(ImportStepTests),
    ))
